"""
For the ``future`` package.

Adds this import line:

    from past.builtins import str as oldstr

at the top and wraps any unadorned string literals 'abc' or explicit byte-string
literals b'abc' in oldstr() calls so the code has the same behaviour on Py3 as
on Py2.6/2.7.
"""

from __future__ import unicode_literals
import re
from lib2to3 import fixer_base
from lib2to3.pgen2 import token
from lib2to3.fixer_util import syms
from libfuturize.fixer_util import (future_import, touch_import_top,
                                    wrap_in_fn_call)


_literal_re = re.compile(r"[^uUrR]?[\'\"]")


class FixOldstrWrap(fixer_base.BaseFix):
    BM_compatible = True
    PATTERN = "STRING"

    def transform(self, node, results):
        if node.type == token.STRING:
            touch_import_top(u'past.types', u'oldstr', node)
            if _literal_re.match(node.value):
                new = node.clone()
                # Strip any leading space or comments:
                # TODO: check: do we really want to do this?
                new.prefix = u''
                new.value = u'b' + new.value
                wrapped = wrap_in_fn_call("oldstr", [new], prefix=node.prefix)
                return wrapped



import sys
from types import CodeType

from . import TemplateSyntaxError
from ._compat import PYPY
from .utils import internal_code
from .utils import missing


def rewrite_traceback_stack(source=None):
    """Rewrite the current exception to replace any tracebacks from
    within compiled template code with tracebacks that look like they
    came from the template source.

    This must be called within an ``except`` block.

    :param exc_info: A :meth:`sys.exc_info` tuple. If not provided,
        the current ``exc_info`` is used.
    :param source: For ``TemplateSyntaxError``, the original source if
        known.
    :return: A :meth:`sys.exc_info` tuple that can be re-raised.
    """
    exc_type, exc_value, tb = sys.exc_info()

    if isinstance(exc_value, TemplateSyntaxError) and not exc_value.translated:
        exc_value.translated = True
        exc_value.source = source

        try:
            # Remove the old traceback on Python 3, otherwise the frames
            # from the compiler still show up.
            exc_value.with_traceback(None)
        except AttributeError:
            pass

        # Outside of runtime, so the frame isn't executing template
        # code, but it still needs to point at the template.
        tb = fake_traceback(
            exc_value, None, exc_value.filename or "<unknown>", exc_value.lineno
        )
    else:
        # Skip the frame for the render function.
        tb = tb.tb_next

    stack = []

    # Build the stack of traceback object, replacing any in template
    # code with the source file and line information.
    while tb is not None:
        # Skip frames decorated with @internalcode. These are internal
        # calls that aren't useful in template debugging output.
        if tb.tb_frame.f_code in internal_code:
            tb = tb.tb_next
            continue

        template = tb.tb_frame.f_globals.get("__jinja_template__")

        if template is not None:
            lineno = template.get_corresponding_lineno(tb.tb_lineno)
            fake_tb = fake_traceback(exc_value, tb, template.filename, lineno)
            stack.append(fake_tb)
        else:
            stack.append(tb)

        tb = tb.tb_next

    tb_next = None

    # Assign tb_next in reverse to avoid circular references.
    for tb in reversed(stack):
        tb_next = tb_set_next(tb, tb_next)

    return exc_type, exc_value, tb_next


def fake_traceback(exc_value, tb, filename, lineno):
    """Produce a new traceback object that looks like it came from the
    template source instead of the compiled code. The filename, line
    number, and location name will point to the template, and the local
    variables will be the current template context.

    :param exc_value: The original exception to be re-raised to create
        the new traceback.
    :param tb: The original traceback to get the local variables and
        code info from.
    :param filename: The template filename.
    :param lineno: The line number in the template source.
    """
    if tb is not None:
        # Replace the real locals with the context that would be
        # available at that point in the template.
        locals = get_template_locals(tb.tb_frame.f_locals)
        locals.pop("__jinja_exception__", None)
    else:
        locals = {}

    globals = {
        "__name__": filename,
        "__file__": filename,
        "__jinja_exception__": exc_value,
    }
    # Raise an exception at the correct line number.
    code = compile("\n" * (lineno - 1) + "raise __jinja_exception__", filename, "exec")

    # Build a new code object that points to the template file and
    # replaces the location with a block name.
    try:
        location = "template"

        if tb is not None:
            function = tb.tb_frame.f_code.co_name

            if function == "root":
                location = "top-level template code"
            elif function.startswith("block_"):
                location = 'block "%s"' % function[6:]

        # Collect arguments for the new code object. CodeType only
        # accepts positional arguments, and arguments were inserted in
        # new Python versions.
        code_args = []

        for attr in (
            "argcount",
            "posonlyargcount",  # Python 3.8
            "kwonlyargcount",  # Python 3
            "nlocals",
            "stacksize",
            "flags",
            "code",  # codestring
            "consts",  # constants
            "names",
            "varnames",
            ("filename", filename),
            ("name", location),
            "firstlineno",
            "lnotab",
            "freevars",
            "cellvars",
        ):
            if isinstance(attr, tuple):
                # Replace with given value.
                code_args.append(attr[1])
                continue

            try:
                # Copy original value if it exists.
                code_args.append(getattr(code, "co_" + attr))
            except AttributeError:
                # Some arguments were added later.
                continue

        code = CodeType(*code_args)
    except Exception:
        # Some environments such as Google App Engine don't support
        # modifying code objects.
        pass

    # Execute the new code, which is guaranteed to raise, and return
    # the new traceback without this frame.
    try:
        exec(code, globals, locals)
    except BaseException:
        return sys.exc_info()[2].tb_next


def get_template_locals(real_locals):
    """Based on the runtime locals, get the context that would be
    available at that point in the template.
    """
    # Start with the current template context.
    ctx = real_locals.get("context")

    if ctx:
        data = ctx.get_all().copy()
    else:
        data = {}

    # Might be in a derived context that only sets local variables
    # rather than pushing a context. Local variables follow the scheme
    # l_depth_name. Find the highest-depth local that has a value for
    # each name.
    local_overrides = {}

    for name, value in real_locals.items():
        if not name.startswith("l_") or value is missing:
            # Not a template variable, or no longer relevant.
            continue

        try:
            _, depth, name = name.split("_", 2)
            depth = int(depth)
        except ValueError:
            continue

        cur_depth = local_overrides.get(name, (-1,))[0]

        if cur_depth < depth:
            local_overrides[name] = (depth, value)

    # Modify the context with any derived context.
    for name, (_, value) in local_overrides.items():
        if value is missing:
            data.pop(name, None)
        else:
            data[name] = value

    return data


if sys.version_info >= (3, 7):
    # tb_next is directly assignable as of Python 3.7
    def tb_set_next(tb, tb_next):
        tb.tb_next = tb_next
        return tb


elif PYPY:
    # PyPy might have special support, and won't work with ctypes.
    try:
        import tputil
    except ImportError:
        # Without tproxy support, use the original traceback.
        def tb_set_next(tb, tb_next):
            return tb

    else:
        # With tproxy support, create a proxy around the traceback that
        # returns the new tb_next.
        def tb_set_next(tb, tb_next):
            def controller(op):
                if op.opname == "__getattribute__" and op.args[0] == "tb_next":
                    return tb_next

                return op.delegate()

            return tputil.make_proxy(controller, obj=tb)


else:
    # Use ctypes to assign tb_next at the C level since it's read-only
    # from Python.
    import ctypes

    class _CTraceback(ctypes.Structure):
        _fields_ = [
            # Extra PyObject slots when compiled with Py_TRACE_REFS.
            ("PyObject_HEAD", ctypes.c_byte * object().__sizeof__()),
            # Only care about tb_next as an object, not a traceback.
            ("tb_next", ctypes.py_object),
        ]

    def tb_set_next(tb, tb_next):
        c_tb = _CTraceback.from_address(id(tb))

        # Clear out the old tb_next.
        if tb.tb_next is not None:
            c_tb_next = ctypes.py_object(tb.tb_next)
            c_tb.tb_next = ctypes.py_object()
            ctypes.pythonapi.Py_DecRef(c_tb_next)

        # Assign the new tb_next.
        if tb_next is not None:
            c_tb_next = ctypes.py_object(tb_next)
            ctypes.pythonapi.Py_IncRef(c_tb_next)
            c_tb.tb_next = c_tb_next

        return tb



# -*- coding: utf-8 -*-

import logging

from application.model import Configurations
from application.model.exceptions import ConfigurationReaderException
from cfgtools import read
from helper import config_logger
from typing import Optional


class Configurator(object):

    def __init__(self, loader_path):
        # type: (str) -> None

        self.loader_path = loader_path
        self.reader = read(self.loader_path)
        self.configs = Configurations()

    def read_loader_configurations(self):
        # type: () -> None

        try:
            self.configs.service_name = str(self.reader.find_value("Identification.ComponentName"))
            self.configs.required_services = "|".join(self._get_config("RequiredServices", array=True) or [])
            self.configs.persistence_name = str(self._get_config("PersistenceName") or "")

        except Exception as ex:
            raise ConfigurationReaderException(ex)

    def configure_logger(self):
        # type: () -> None

        config_logger(self.loader_path, self.configs.service_name)
        self.configs.logger = logging.getLogger(self.configs.service_name)

    def _get_config(self, key, array=False, group=None):
        # type: (str, Optional[bool], str) -> str
        group_name = self.configs.service_name if group is None else group
        path = "{}.{}".format(group_name, key)
        return self.reader.find_values(path) if array else self.reader.find_value(path)



from application.service import BenefitService
from application.model.transactions import AddBenefitToApplyUnitOfWork


class AddBenefitToApplyInteractor:
    def __init__(self, benefit_service):
        # type: (BenefitService) -> None

        self.benefit_service = benefit_service

    def apply(self, uow):
        # type: (AddBenefitToApplyUnitOfWork) -> None

        self.benefit_service.apply_benefit(uow, is_manual=True)



import os
import sys

py3 = sys.version_info.major >= 3
py36 = sys.version_info[0:2] >= (3, 6)

if py3:

    def isStringType(t):
        return isinstance(t, str)

    if py36:
        from pathlib import Path

        def isPath(f):
            return isinstance(f, (bytes, str, Path))

    else:

        def isPath(f):
            return isinstance(f, (bytes, str))


else:

    def isStringType(t):
        return isinstance(t, basestring)  # noqa: F821

    def isPath(f):
        return isinstance(f, basestring)  # noqa: F821


# Checks if an object is a string, and that it points to a directory.
def isDirectory(f):
    return isPath(f) and os.path.isdir(f)


class deferred_error(object):
    def __init__(self, ex):
        self.ex = ex

    def __getattr__(self, elt):
        raise self.ex



class Table(object):
    def __init__(
            self,                       # type: str
            id="",                      # type: str
            name="",                    # type: str
            current_pos_id="",          # type: str
            status="",                  # type: str
            tab_id="",                  # type: str
            business_period="",         # type: str
            sector="",                  # type: str
            slice_parent_table_id="",   # type: str
            slice_source_table_id="",   # type: str
            mode=""                     # type: str
    ):
        # type: (...) -> None

        self.id = id
        self.name = name
        self.current_pos_id = current_pos_id
        self.status = status
        self.tab_id = tab_id
        self.business_period = business_period
        self.sector = sector
        self.slice_source_table_id = slice_source_table_id
        self.slice_parent_table_id = slice_parent_table_id
        self.mode = mode



# -*- coding: utf-8 -*-
import datetime

from application.interactor import OpenDayInteractor
from msgbus import TK_POS_BUSINESSBEGIN, FM_PARAM, TK_SYS_ACK, TK_PERSIST_SEQUENCER_RESET
from sysactions import get_model, format_date, get_business_period,  translate_message, sys_log_info, send_message, \
     sys_log_exception, print_report, action, is_day_opened, StopAction, is_day_blocked, show_messagebox
from actions.util.get_menu_user_logged import get_menu_manager_authenticated_user

USECS_PER_SEC = 1000000

poslist = []

nf_type = None
mbcontext = None

BUTTONS = "$OK|$CANCEL"


@action
def openday(pos_id, store_wide="false", force_open="false", *args):
    model = get_model(pos_id)
    if store_wide == "false":
        check_business_day(pos_id, model=model, need_opened=False)

    OpenDayInteractor().check_has_update_menu(pos_id)

    posnumbers = tuple(poslist) if (store_wide.lower() == "true") else (pos_id,)

    cannot_open = []
    poslist_new = []

    for x in posnumbers:
        model_x = get_model(x)
        x_period = get_business_period(model_x)

        if model_x.find("PosState").get("state") not in ("CLOSED", "UNDEFINED"):
            cannot_open.append(translate_message(model_x, "DAY_OPEN_NOT_CLOSED", "%02d" % int(x), format_date(x_period)))
            continue

        poslist_new.append(x)

    if len(cannot_open) > 0:
        pos_nok = ""
        for opt in sorted(cannot_open):
            pos_nok += "%s\\" % opt

        message = ("$REGISTERS_NOT_OPENED_CONFIRM|%s" % pos_nok) if poslist_new else ("$REGISTERS_NOT_OPENED|%s" % pos_nok)
        options = ("$CANCEL", "$OPEN_OTHER_REGISTERS") if poslist_new else ("$CANCEL",)
        icon = "question" if poslist_new else "error"
        question = show_messagebox(pos_id, message=message, title="$DAY_OPEN", icon=icon, buttons="|".join(options))
        if (question is None) or (question == 0):
            return  # User cancelled, or timeout

        posnumbers = poslist_new

    posnumbers = sorted(posnumbers)
    sys_log_info("Opening business day. [Store-wide: %s] posnumbers: %s" % (store_wide, posnumbers))
    pos_ok = []

    today = datetime.date.today().strftime("%Y%m%d")
    for posno in posnumbers:
        try:
            model = get_model(posno)
            sys_log_info("Opening business day on POS id: [%s]" % posno)
            msg = send_message("POS%d" % int(posno), TK_POS_BUSINESSBEGIN, FM_PARAM, "%s\0%s" % (posno, today), timeout=3600 * USECS_PER_SEC)
            if msg.token == TK_SYS_ACK:
                pos_ok.append(posno)
                continue
            else:
                sys_log_exception("Error opening business day on pos id: %s" % posno)
                errors = msg.data.split('\0')
                errormsg = errors[1] if len(errors) > 1 else errors[0] if len(errors) > 0 else errors
                show_messagebox(pos_id, "$ERROR_OPENING_BUSINESS_PERIOD|%02d|%s" % (int(posno), errormsg))

        except Exception as ex:
            sys_log_exception("Error opening business day on pos id: %s" % posno)
            show_messagebox(pos_id, "$ERROR_OPENING_BUSINESS_PERIOD|%02d|%s" % (int(posno), str(ex)))

    posnumbers = "|".join(map(str, pos_ok))
    if len(pos_ok):
        menu_manager_user = get_menu_manager_authenticated_user(int(pos_id))
        print_report(pos_id, model, True, "dayOpen_report", pos_id, today, posnumbers, store_wide, menu_manager_user)
    # Reset daily counters
    send_message("Persistence", TK_PERSIST_SEQUENCER_RESET, format=FM_PARAM, data="TraySetNameGen")


def check_business_day(posid, model=None, need_opened=True, can_be_blocked=True):
    if model is None:
        model = get_model(posid)
    opened = is_day_opened(model)
    if opened and not need_opened:
        show_messagebox(posid, "$NEED_TO_CLOSE_BDAY_FIRST")
        raise StopAction()
    if need_opened and not opened:
        show_messagebox(posid, "$NEED_TO_OPEN_BDAY_FIRST")
        raise StopAction()
    if (not can_be_blocked) and is_day_blocked(model):
        show_messagebox(posid, "$POS_IS_BLOCKED_BY_TIME")
        raise StopAction()



class OrderInsertPaymentResponse(object):
    def __init__(self, tender):
        # type: (dict) -> None

        self.dueAmount = tender.get("dueAmount")
        self.tenderId = tender.get("tenderId")



# -*- coding: utf-8 -*-
import json
from datetime import datetime


from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection

from ..model import SaleChallengeItem


class SaleChallengeRepository(BaseRepository):

    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None
        super(SaleChallengeRepository, self).__init__(mb_context)

    def get_sale_challenge(self, sale_challenge_id):
        # type: (int) -> SaleChallengeItem

        def inner_func(conn):
            # type: (Connection) -> SaleChallengeItem

            query = self._SaleChallengeProductsQuery
            cursor = conn.select(query.format(sale_challenge_id))
            row = cursor.get_row(0)
            return SaleChallengeItem(row.get_entry("Name"),
                                     datetime.strptime(row.get_entry("StartDate"), "%Y-%m-%d"),
                                     datetime.strptime(row.get_entry("EndDate"), "%Y-%m-%d"),
                                     json.loads(row.get_entry("Products") or "[]"))
        return self.execute_with_connection(inner_func, service_name="DailyGoalsPersistence")

    _SaleChallengeProductsQuery = """\
        SELECT *
        FROM SalesChallenge
        WHERE Id = {0}
    """



from _CloseDaySummaryInfo import CloseDaySummaryInfo
from _CloseDaySummaryDto import CloseDaySummaryDto
from _CloseDaySummaryHeaderDto import CloseDaySummaryHeaderDto
from _CloseDaySummaryBodyDto import CloseDaySummaryBodyDto
from _CashReportBodyDto import CashReportBodyDto
from _CashReportBodyLineDto import CashReportBodyLineDto
from _OrderInfo import OrderInfo

__all__ = ["CloseDaySummaryInfo",
           "CloseDaySummaryDto",
           "CloseDaySummaryHeaderDto",
           "CloseDaySummaryBodyDto",
           "CashReportBodyDto",
           "CashReportBodyLineDto",
           "OrderInfo"]



class Route(object):
    def __init__(self, method, path):
        self.method = method
        self.path = path

    def __eq__(self, o):
        if not isinstance(o, Route):
            return False

        return self.method == o.method and self.path == o.path

    def __ne__(self, o):
        return not self.__eq__(o)

    def __hash__(self):
        return hash((self.method, self.path))

    def __str__(self):
        return "method: {}, path: {}".format(self.method, self.path)

    def __repr__(self):
        return "Route(\"{}\", \"{}\")".format(self.method, self.path)



# encoding: utf-8

from typing import Dict, Optional, Tuple, List

from benefitsreport.commons import discount_in_all_items, check_item_in_action
from benefitsreport.model import OrderBenefit
from benefitsreport.repository import BenefitsOrderRepository


class BenefitsReport(object):
    def __init__(self, order_repository):
        # type: (BenefitsOrderRepository) -> None

        self.order_repository = order_repository

    def get_benefits_orders_detailed(self, initial_date, end_date, operator_id, benefits):
        # type: (str, str, str, Dict) -> Dict

        orders = self.order_repository.get_orders_benefit_detailed(initial_date, end_date, operator_id, benefits)
        result = {}
        for period in list(dict.fromkeys([item.business_period for item in orders])):
            filtered_orders = filter(lambda x: self._filter_item(x, period), orders)
            if len(filtered_orders) > 0:
                result[period] = filtered_orders

        return result

    def get_benefits_orders_consolidated(self, initial_date, end_date, operator_id):
        # type: (str, str, str) -> Dict

        orders = self.order_repository.get_orders_benefit_consolidated(initial_date, end_date, operator_id)
        benefits_keys = [item.applied_benefits.benefit_name for item in orders]
        result = {}
        for benefits_key in list(dict.fromkeys(benefits_keys)):
            items = filter(lambda x: self._filter_item_consolidated(x, benefits_key), orders)
            if len(items) > 0:
                values = [self._get_data_collum(line) for line in items]
                result[str(benefits_key)] = self._sum_vales(values)

        return result

    def _filter_item(self, item, period):
        # type: (OrderBenefit, str) -> Optional[OrderBenefit]

        if item.business_period != period:
            return
        item = self._get_discount_item(item)
        if item.discount_value > 0:
            return item

    @staticmethod
    def _filter_item_consolidated(item, benefits_key):
        # type: (OrderBenefit, str) -> Optional[OrderBenefit]

        benefit_total_value = item.applied_benefits.benefit_total_value
        if item.applied_benefits.benefit_name == benefits_key and benefit_total_value > 0:
            discount_items = item.applied_benefits.discount_item_values
            line_number = item.line_number
            items = [item for key in discount_items.keys() if line_number in discount_items[key].keys()]
            if len(items) == 0:
                return

            return item

    def _get_discount_item(self, item):
        # type: (OrderBenefit) -> Optional[OrderBenefit]

        try:
            self._filter_promotion_tender(item)
            discount_items = item.applied_benefits.discount_item_values
            for _, value in discount_items.iteritems():
                if item.line_number in value.keys():
                    value_discount = value[item.line_number]
                    item.discount_value += item.ordered_qty * value_discount
        except KeyError as _:
            pass
        return item

    def _filter_promotion_tender(self, item):
        # type: (OrderBenefit) -> OrderBenefit

        try:
            burnt_items = item.applied_benefits.burnt_items
            item_burn = burnt_items[item.key_burn]
            if item_burn < 0:
                raise KeyError()

            item.discount_value = sum([float(x.get("amount")) for x in item.applied_benefits.promotion_tender])
        except KeyError as _:
            item = self._check_item_in_other_line(item)
        return item

    @staticmethod
    def _check_item_in_other_line(item):
        # type: (OrderBenefit) -> OrderBenefit

        benefit = item.applied_benefits.benefit
        part_code = item.part_code
        discount_all_items = discount_in_all_items(benefit)
        if not discount_all_items and check_item_in_action(benefit, part_code):
            promotion_tender = item.applied_benefits.promotion_tender
            item.discount_value = sum([float(x.get("amount")) for x in promotion_tender])

        return item

    @staticmethod
    def _sum_vales(values):
        # type: (List) -> Tuple

        qty = 0
        default_value = 0.0
        discount_value = 0.0
        for value in list(values):
            qty += value[0]
            default_value += value[1] if value[0] == 1 else value[0] * value[1]
            discount_value += value[2]

        benefit_description = values[0][3]
        return qty, default_value, discount_value, benefit_description

    def _get_data_collum(self, item):
        # type: (OrderBenefit) -> List

        benefit_applied = item.applied_benefits
        benefit_description = benefit_applied.benefit_description
        item = self._get_discount_item(item)
        return [item.ordered_qty, item.default_unit_price, item.discount_value, benefit_description]



# -*- coding: utf-8 -*-
from kdsmonitor import KdsMonitor
from messagehandler import EventHandler
from msgbus import TK_SYS_NAK, TK_SYS_ACK, MBEasyContext, MBException, TK_KDS_GETMODEL, FM_PARAM
from bustoken import TK_KDSMONITOR_STATUS_REQUEST
from xml.etree import cElementTree as eTree
import ast


class KdsMonitorEventHandler(EventHandler):
    def __init__(self, mbcontext, kds_monitor_thread):
        # type: (MBEasyContext, KdsMonitor) -> None
        super(KdsMonitorEventHandler, self).__init__(mbcontext)
        self.kds_monitor_thread = kds_monitor_thread

    def get_handled_tokens(self):
        return [TK_KDSMONITOR_STATUS_REQUEST]

    def handle_message(self, msg):
        try:
            if msg.token == TK_KDSMONITOR_STATUS_REQUEST:
                response = self._handle_status_request(msg)
            else:
                response = (False, "")

            msg.token = TK_SYS_ACK if response[0] else TK_SYS_NAK
            self.mbcontext.MB_ReplyMessage(msg, data=response[1])

        except Exception as ex:
            # sys_log_exception("Unexpected exception when handling event {0}. Error message: {1}".format(msg.token, ex.message))
            msg.token = TK_SYS_NAK
            self.mbcontext.MB_ReplyMessage(msg, data=str(ex))

        return False

    def handle_event(self, subject, evt_type, data, msg):
        raise NotImplementedError()

    def _handle_status_request(self, msg):
        try:
            model_msg = self.mbcontext.MB_EasySendMessage("KdsController", TK_KDS_GETMODEL, format=FM_PARAM, data="1", timeout=10000 * 1000)
            model_tree = eTree.XML(model_msg.data)
            kdss_elements = model_tree.find("KDSs").findall('KDS')
            kds_status_list = []
            kds_info = []
            for kds in self.kds_monitor_thread.kdss_to_monitor:
                kds_id = (str(kds.name)).strip("KDS")
                for kds_element in kdss_elements:
                    kds_element_id = kds_element.get('id')
                    if kds_id == kds_element_id:
                        kds_info.append(kds_id)
                        kds_info.append(kds_element.get('desc'))
                        if kds.enabled:
                            kds_info.append("Enabled")
                        else:
                            kds_info.append("Disabled")
                        kds_status_list.append(kds_info)
                        kds_info = []
                        break
            return True, str(kds_status_list)
        except MBException as ex:
            return False, "Erro ao procurar KDSs"

    def terminate_event(self):
        self.kds_monitor_thread.finish()



# -*- coding: utf-8 -*-
import json
from xml.etree import cElementTree as eTree

from actions.delivery import is_delivery_fee_enabled
from actions.delivery.get_delivery_fee_code import get_delivery_fee_code
from old_helper import round_half_away_from_zero
from sysactions import action, get_model, get_posot, get_operator_session, StopAction, translate_message, \
    show_confirmation, get_current_order

from bustoken import TK_REMOTE_ORDER_SEND_ORDER_TO_PRODUCTION
from msgbus import FM_PARAM, TK_SYS_ACK
from sysactions import show_messagebox

from actions.functions.do_store_order import doStoreOrder
from model.customexception import SendOrderToProductionError
from .. import logger
from .. import mb_context


@action
def produce_delivery_order(pos_id, order_id, manual_recall='false', warn_inconsistencies='false'):
    model = get_model(pos_id)
    pos_ot = get_posot(model)
    session = get_operator_session(model)

    if not _validate_manual_order(model, pos_ot, pos_id, order_id, warn_inconsistencies.lower() != 'false'):
        return

    try:
        manual_recall = manual_recall.lower() == 'true'
        if manual_recall:
            doStoreOrder(pos_id, show_popup=False)

        msg = mb_context.MB_EasySendMessage("RemoteOrder", TK_REMOTE_ORDER_SEND_ORDER_TO_PRODUCTION, FM_PARAM, order_id)
        if msg.token == TK_SYS_ACK:
            show_messagebox(pos_id, "$ORDER_PRODUCED")
        else:
            raise SendOrderToProductionError("$ERROR_SENDING_REMOTE_ORDER_TO_PRODUCTION|{}".format(msg.data))
    except Exception as ex:
        logger.exception("Error producing delivery order")
        show_messagebox(pos_id, message=ex.message)

        if manual_recall:
            pos_ot.recallOrder(int(pos_id), int(order_id), session)

        raise StopAction()


def _validate_manual_order(model, pos_ot, pos_id, order_id, warn_inconsistencies):
    try:
        order_picture = eTree.XML(pos_ot.orderPicture(pos_id, order_id))
        manual_order = order_picture.find("Order")
        remote_order_json_text = order_picture.find(".//OrderProperty[@key='REMOTE_ORDER_JSON']")
        remote_order_json_text = remote_order_json_text.get("value", None) if remote_order_json_text else None
        if not remote_order_json_text or not warn_inconsistencies:
            return True

        remote_json = json.loads(remote_order_json_text)

        order_items_quantity = _get_order_items_quantity(manual_order)
        order_total_value = _get_order_total_value(manual_order)
        json_items_quantity = _get_json_items_quantity(remote_json)
        json_total_value = _get_json_total_value(remote_json)

        quantity_is_different, quantity_diff = _items_quantity_is_different(order_items_quantity, json_items_quantity)
        value_is_different, value_diff = _value_is_different(order_total_value, json_total_value)

        if quantity_is_different or value_is_different:
            _confirm_order_difference(model, pos_id, remote_json)
            if quantity_is_different:
                _validate_quantity_difference(model, pos_id, order_items_quantity, json_items_quantity)
            if value_is_different:
                _validate_value_difference(model, pos_id, value_diff, order_total_value, json_total_value)

    except StopAction:
        return False
    except Exception as _: # noqa
        logger.exception("Error validating manual order")
        return False

    return True


def _get_json_total_discounts_values(remote_json):
    discounts = {}
    vouchers = remote_json.get("vouchers")
    if not vouchers:
        return discounts
    store_discount = vouchers.get("store")
    if not store_discount:
        return discounts
    return store_discount


def _validate_value_difference(model, pos_id, value_diff, order_total_value, json_total_value):
    message = "$CONFIRM_MANUAL_ORDER_DISCOUNT" if value_diff > 0 else "$CONFIRM_MANUAL_ORDER_ADDITION"
    message = message + "|{0}|{1}|{2}".format(_format_currency(model, order_total_value),
                                              _format_currency(model, json_total_value),
                                              _format_currency(model, abs(value_diff)))
    if not show_confirmation(pos_id, message):
        raise StopAction()

    _apply_difference_discount(pos_id, value_diff)


def _validate_quantity_difference(model, pos_id, order_items_quantity, json_items_quantity):
    message = "$CONFIRM_MANUAL_ORDER_ITEM_QUANTITY_DIFF"
    message = message + "|{0}|{1}".format(order_items_quantity, json_items_quantity)
    if not show_confirmation(pos_id, message):
        raise StopAction()


def _apply_difference_discount(pos_id, value_diff):
    from actions.discount.apply_discount import apply_amount_discount

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order = get_current_order(model)
    apply_amount_discount(pos_id, pos_ot, order, discount_type=1, discount_value=float(value_diff))


def _confirm_order_difference(model, pos_id, remote_json):
    partner_order_id = _get_json_order_id(remote_json)
    message = translate_message(model, "DELIVERY_RECALL_WARNING")
    message = message.format(
        partner_order_id,
        remote_json.get("partner", "").upper(),
        str(_get_json_items_quantity(remote_json)),
        _format_currency(model, _get_json_total_value(remote_json))
    )
    if not show_confirmation(pos_id, message):
        raise StopAction()


def _format_currency(model, value):
    currency_symbol = translate_message(model, "L10N_CURRENCY_SYMBOL")
    return "{0} {1:.2f}".format(currency_symbol, value)


def _items_quantity_is_different(order_items_quantity, json_items_quantity):
    is_different = not (order_items_quantity == json_items_quantity)
    value_diff = order_items_quantity - json_items_quantity

    return is_different, value_diff


def _value_is_different(order_total_value, json_total_value):
    round_order_value = round_half_away_from_zero(order_total_value, 2)
    round_json_value = round_half_away_from_zero(json_total_value, 2)
    is_different = not (round_order_value == round_json_value)
    value_diff = round_half_away_from_zero(order_total_value - json_total_value, 2)

    return is_different, value_diff


def _get_order_total_value(manual_order):
    manual_order_total = float(manual_order.get("totalAfterDiscount", 0)) + float(manual_order.get("taxTotal", 0))
    return manual_order_total


def _get_json_total_value(remote_json):
    json_discounts = _get_json_total_discounts_values(remote_json)

    json_order_discount = json_discounts.get("order", 0)
    json_total_value = float(remote_json.get("subTotal", 0)) - json_order_discount

    delivery_fee_value = 0
    if is_delivery_fee_enabled():
        json_delivery_fee_discount = json_discounts.get("delivery_fee", 0)
        delivery_fee_value = float(remote_json.get("deliveryFee", 0)) - json_delivery_fee_discount

    return json_total_value + delivery_fee_value


def _get_order_items_quantity(manual_order):
    sale_lines = manual_order.findall(".//SaleLine[@level='0']")
    delivery_fee_code = get_delivery_fee_code()

    return sum(float(c.get("qty", "0")) for c in _get_valid_sale_lines(sale_lines, delivery_fee_code))


def _get_valid_sale_lines(sale_lines, fee_code):
    return [line for line in sale_lines if (float(line.get("qty", "0")) > 0 and line.get("partCode", "") != fee_code)]


def _get_json_items_quantity(remote_json):
    valid_items = [line for line in remote_json.get("items", []) if (float(line.get("quantity", "0")) > 0)]

    return sum(float(c.get("quantity", "0")) for c in valid_items)


def _get_json_order_id(remote_json):
    return remote_json.get("shortReference", "")



from __future__ import absolute_import, division, print_function

from ._core import (
    AbstractPEMObject,
    Certificate,
    CertificateRequest,
    CertificateRevocationList,
    DHParameters,
    ECPrivateKey,
    Key,
    OpenSSHPrivateKey,
    PrivateKey,
    PublicKey,
    RSAPrivateKey,
    RSAPublicKey,
    parse,
    parse_file,
)


try:
    from . import twisted
except ImportError:
    twisted = None  # type: ignore


__version__ = "19.3.0"
__author__ = "Hynek Schlawack"
__license__ = "MIT"
__description__ = "Easy PEM file parsing in Python."
__url__ = "https://pem.readthedocs.io/"
__uri__ = __url__
__email__ = "hs@ox.cx"


__all__ = [
    "AbstractPEMObject",
    "Certificate",
    "CertificateRequest",
    "CertificateRevocationList",
    "DHParameters",
    "ECPrivateKey",
    "Key",
    "OpenSSHPrivateKey",
    "PrivateKey",
    "PublicKey",
    "RSAPrivateKey",
    "RSAPublicKey",
    "parse",
    "parse_file",
    "twisted",
]



from abc import ABCMeta, abstractmethod
from xml.etree import cElementTree as eTree

from model.order_detail import OrderDetail


class OrderServiceModel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def convert_mw_order_to_order_detail(self, order):
        # type: (eTree) -> OrderDetail

        raise NotImplementedError



"""
This disables builtin functions (and one exception class) which are
removed from Python 3.3.

This module is designed to be used like this::

    from future.builtins.disabled import *

This disables the following obsolete Py2 builtin functions::

    apply, cmp, coerce, execfile, file, input, long,
    raw_input, reduce, reload, unicode, xrange

We don't hack __builtin__, which is very fragile because it contaminates
imported modules too. Instead, we just create new functions with
the same names as the obsolete builtins from Python 2 which raise
NameError exceptions when called.

Note that both ``input()`` and ``raw_input()`` are among the disabled
functions (in this module). Although ``input()`` exists as a builtin in
Python 3, the Python 2 ``input()`` builtin is unsafe to use because it
can lead to shell injection. Therefore we shadow it by default upon ``from
future.builtins.disabled import *``, in case someone forgets to import our
replacement ``input()`` somehow and expects Python 3 semantics.

See the ``future.builtins.misc`` module for a working version of
``input`` with Python 3 semantics.

(Note that callable() is not among the functions disabled; this was
reintroduced into Python 3.2.)

This exception class is also disabled:

    StandardError

"""

from __future__ import division, absolute_import, print_function

from future import utils


OBSOLETE_BUILTINS = ['apply', 'chr', 'cmp', 'coerce', 'execfile', 'file',
                     'input', 'long', 'raw_input', 'reduce', 'reload',
                     'unicode', 'xrange', 'StandardError']


def disabled_function(name):
    '''
    Returns a function that cannot be called
    '''
    def disabled(*args, **kwargs):
        '''
        A function disabled by the ``future`` module. This function is
        no longer a builtin in Python 3.
        '''
        raise NameError('obsolete Python 2 builtin {0} is disabled'.format(name))
    return disabled


if not utils.PY3:
    for fname in OBSOLETE_BUILTINS:
        locals()[fname] = disabled_function(fname)
    __all__ = OBSOLETE_BUILTINS
else:
    __all__ = []



from xml.etree import ElementTree as eTree
from common import TipBuilderCommon
from helper import remove_accents
from typing import Dict


class PrinterCommon(object):
    def __init__(self, tip_config, tip_config_range, tip_builder_common):
        # type: (Dict, Dict, TipBuilderCommon) -> None

        self.tip_config = tip_config
        self.tip_config_range = tip_config_range
        self.tip_builder_common = tip_builder_common

    def get_tip_order(
        self,
        order,              # type: eTree
        is_mobile=False     # type: bool
    ):
        # type: (...) -> str

        if len(self.tip_config_range) == 0 and len(self.tip_config) == 0:
            return ""

        order_tip = float(order.get("tip"))
        if len(self.tip_config_range) == 0 and len(self.tip_config) > 0:
            return self._get_tip_line(self.tip_config['TipCode'], self.tip_config['TipName'], order_tip, is_mobile)

        report = ""

        total_gross = float(order.get('totalGross'))
        tip_ranges = self.tip_builder_common.get_percentage_by_tip(total_gross, order_tip)
        for key, value in tip_ranges.items():
            report += self._get_tip_line(key, value.get("TipName"), value.get("TipValue"), is_mobile)

        return report

    @staticmethod
    def value_tip_by_percentage(percent, order_value):
        # type: (float, float) -> float

        return (percent * order_value) / 100.0

    @staticmethod
    def _get_tip_line(
        code,           # type: str
        description,    # type: str
        value,          # type: float
        is_mobile       # type: bool
    ):
        # type: (...) -> str

        format_mobile = "%-7.7s %-14.14s%6.3f%-3.2s%9.2f"
        format_desktop = "%-7.7s %-18.18s%6.3f%2.2s%9.02f"
        order_tip_format = format_mobile if is_mobile else format_desktop

        tip_line = remove_accents(order_tip_format % (code, description, float(1), "un", value))

        format_mobile = "%-39.40s %7.02f\n"
        format_desktop = "%-46.46s %9.02f\n"
        order_tip_format = format_mobile if is_mobile else format_desktop

        return order_tip_format % (tip_line, value)



# -*- coding: utf-8 -*-

import sysactions

from bustoken import TK_FISCALWRAPPER_CANCEL_ORDER
from msgbus import FM_PARAM, TK_SYS_ACK
from mw_helper import show_filtered_list_box_dialog
from typing import List

from .. import mb_context


def void_order(pos_id, last_paid_order="0", order_id='', abandon=0, void_reason=9, pos_ot=None):
    model = sysactions.get_model(pos_id)
    if pos_ot is None:
        pos_ot = sysactions.get_posot(model)

    order_id = int(order_id) if order_id != "" else ""
    void_order_id = order_id
    if (type(void_reason) == int) or not void_reason:
        void_reason = get_void_reason_id(pos_id, void_reason)
    if not order_id:
        order = sysactions.get_current_order(model)
        order_id = order.get('orderId')

    pos_ot.setOrderCustomProperties(void_reason, str(order_id))
    pos_ot.voidOrder(int(pos_id), int(last_paid_order), void_order_id, abandon)

    return True


def void_fiscal_order(pos_id, order_id):
    data = "|".join([str(pos_id), str(order_id)])
    msg = mb_context.MB_EasySendMessage("FiscalWrapper", TK_FISCALWRAPPER_CANCEL_ORDER, format=FM_PARAM, data=data)
    if msg.token != TK_SYS_ACK:
        sysactions.show_messagebox(pos_id, message="$ERROR_CANCEL_FISCAL_ORDER|{}".format(msg.data))
        raise sysactions.StopAction(msg.data)


def get_void_reason_id(pos_id, selected_index=None):

    void_reason_ids = range(1, 13)
    void_reason_options = get_void_reason_options(pos_id)
    void_reason_to_show = void_reason_options[0:4]

    if not selected_index:
        message = "$SELECT_REASON"
        selected_index = show_filtered_list_box_dialog(pos_id, void_reason_to_show, message, mask="NOFILTER")

    if selected_index is None:
        return None

    void_reason_id = void_reason_ids[int(selected_index)]
    void_reason_description = void_reason_options[int(selected_index)]
    void_reason_dict = {'VOID_REASON_ID': void_reason_id, 'VOID_REASON_DESCR': void_reason_description}

    return void_reason_dict


def get_void_reason_options(
    pos_id  # type: str
):
    # type: (...) -> List[str]

    void_reason_options = ["CHANGED_YOUR_MIND",
                           "DUPLICATE",
                           "WRONG_SALE",
                           "CANCELLATION",
                           "COUPON_CANCELED",
                           "PAYMENT_CANCELED",
                           "SAVED_ORDER_CANCELED",
                           "ERROR_CREATING_SALE",
                           "ERROR_PROCESSING_PAYMENT_SELECTION",
                           "CANCELLATION_VIA_SYSTEM",
                           "SYNC_ERROR",
                           "MANAGER_AUTHORIZATION_CANCELLATION"]

    return _translate_array(pos_id, void_reason_options)


def _translate_array(pos_id, messages):
    # type: (str, List[str]) -> List[str]

    model = sysactions.get_model(pos_id)
    result = []
    for index, message in enumerate(messages, start=1):
        result.append("{} - {}".format(index, sysactions.translate_message(model, message)))

    return result



# For backwards compatibility, provide imports that used to be here.
from .connection import is_connection_dropped
from .request import make_headers
from .response import is_fp_closed
from .ssl_ import (
    SSLContext,
    HAS_SNI,
    assert_fingerprint,
    resolve_cert_reqs,
    resolve_ssl_version,
    ssl_wrap_socket,
)
from .timeout import (
    current_time,
    Timeout,
)

from .retry import Retry
from .url import (
    get_host,
    parse_url,
    split_first,
    Url,
)



from logging import Logger

from model.error_type import PosValidationErrorsType
from interactor.route.order.reopen_shopping_cart.dto import ReopenShoppingCartRequest
from model.config import Config
from model.exception import ValidationError
from model.order_detail import OrderDetail
from model.repository import PosRepositoryModel
from model.services import OrderServiceModel
from validation_util import \
    check_if_order_is_totaled, \
    check_if_business_day_is_opened, \
    check_if_operator_is_logged, validate_api_key


class ReopenShoppingCartInteractor(object):
    def __init__(
            self,
            logger,             # type: Logger
            config,             # type: Config
            pos_repository,     # type: PosRepositoryModel
            order_service       # type: OrderServiceModel
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.order_service = order_service

    def execute(self, request):
        # type: (ReopenShoppingCartRequest) -> OrderDetail
        pos_id = validate_api_key(request, self.config.api_key_mapping)
        if not pos_id:
            raise ValidationError(PosValidationErrorsType.PosIdNotFound)

        model = self.pos_repository.get_model(pos_id)

        check_if_business_day_is_opened(self.pos_repository, model)
        check_if_operator_is_logged(self.pos_repository, model)
        check_if_order_is_totaled(self.pos_repository, model)

        pos_ot = self.pos_repository.get_pos_ot(model)
        order = self.pos_repository.reopen_shopping_cart(
            pos_id=pos_id,
            pos_ot=pos_ot
        )

        order_detail = self.order_service.convert_mw_order_to_order_detail(order)

        return order_detail



# -*- coding: utf-8 -*-
"""
flask.json
~~~~~~~~~~

:copyright: 2010 Pallets
:license: BSD-3-Clause
"""
import codecs
import io
import uuid
from datetime import date
from datetime import datetime

from itsdangerous import json as _json
from jinja2 import Markup
from werkzeug.http import http_date

from .._compat import PY2
from .._compat import text_type
from ..globals import current_app
from ..globals import request

try:
    import dataclasses
except ImportError:
    dataclasses = None

# Figure out if simplejson escapes slashes.  This behavior was changed
# from one version to another without reason.
_slash_escape = "\\/" not in _json.dumps("/")


__all__ = [
    "dump",
    "dumps",
    "load",
    "loads",
    "htmlsafe_dump",
    "htmlsafe_dumps",
    "JSONDecoder",
    "JSONEncoder",
    "jsonify",
]


def _wrap_reader_for_text(fp, encoding):
    if isinstance(fp.read(0), bytes):
        fp = io.TextIOWrapper(io.BufferedReader(fp), encoding)
    return fp


def _wrap_writer_for_text(fp, encoding):
    try:
        fp.write("")
    except TypeError:
        fp = io.TextIOWrapper(fp, encoding)
    return fp


class JSONEncoder(_json.JSONEncoder):
    """The default Flask JSON encoder. This one extends the default
    encoder by also supporting ``datetime``, ``UUID``, ``dataclasses``,
    and ``Markup`` objects.

    ``datetime`` objects are serialized as RFC 822 datetime strings.
    This is the same as the HTTP date format.

    In order to support more data types, override the :meth:`default`
    method.
    """

    def default(self, o):
        """Implement this method in a subclass such that it returns a
        serializable object for ``o``, or calls the base implementation (to
        raise a :exc:`TypeError`).

        For example, to support arbitrary iterators, you could implement
        default like this::

            def default(self, o):
                try:
                    iterable = iter(o)
                except TypeError:
                    pass
                else:
                    return list(iterable)
                return JSONEncoder.default(self, o)
        """
        if isinstance(o, datetime):
            return http_date(o.utctimetuple())
        if isinstance(o, date):
            return http_date(o.timetuple())
        if isinstance(o, uuid.UUID):
            return str(o)
        if dataclasses and dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if hasattr(o, "__html__"):
            return text_type(o.__html__())
        return _json.JSONEncoder.default(self, o)


class JSONDecoder(_json.JSONDecoder):
    """The default JSON decoder.  This one does not change the behavior from
    the default simplejson decoder.  Consult the :mod:`json` documentation
    for more information.  This decoder is not only used for the load
    functions of this module but also :attr:`~flask.Request`.
    """


def _dump_arg_defaults(kwargs, app=None):
    """Inject default arguments for dump functions."""
    if app is None:
        app = current_app

    if app:
        bp = app.blueprints.get(request.blueprint) if request else None
        kwargs.setdefault(
            "cls", bp.json_encoder if bp and bp.json_encoder else app.json_encoder
        )

        if not app.config["JSON_AS_ASCII"]:
            kwargs.setdefault("ensure_ascii", False)

        kwargs.setdefault("sort_keys", app.config["JSON_SORT_KEYS"])
    else:
        kwargs.setdefault("sort_keys", True)
        kwargs.setdefault("cls", JSONEncoder)


def _load_arg_defaults(kwargs, app=None):
    """Inject default arguments for load functions."""
    if app is None:
        app = current_app

    if app:
        bp = app.blueprints.get(request.blueprint) if request else None
        kwargs.setdefault(
            "cls", bp.json_decoder if bp and bp.json_decoder else app.json_decoder
        )
    else:
        kwargs.setdefault("cls", JSONDecoder)


def detect_encoding(data):
    """Detect which UTF codec was used to encode the given bytes.

    The latest JSON standard (:rfc:`8259`) suggests that only UTF-8 is
    accepted. Older documents allowed 8, 16, or 32. 16 and 32 can be big
    or little endian. Some editors or libraries may prepend a BOM.

    :param data: Bytes in unknown UTF encoding.
    :return: UTF encoding name
    """
    head = data[:4]

    if head[:3] == codecs.BOM_UTF8:
        return "utf-8-sig"

    if b"\x00" not in head:
        return "utf-8"

    if head in (codecs.BOM_UTF32_BE, codecs.BOM_UTF32_LE):
        return "utf-32"

    if head[:2] in (codecs.BOM_UTF16_BE, codecs.BOM_UTF16_LE):
        return "utf-16"

    if len(head) == 4:
        if head[:3] == b"\x00\x00\x00":
            return "utf-32-be"

        if head[::2] == b"\x00\x00":
            return "utf-16-be"

        if head[1:] == b"\x00\x00\x00":
            return "utf-32-le"

        if head[1::2] == b"\x00\x00":
            return "utf-16-le"

    if len(head) == 2:
        return "utf-16-be" if head.startswith(b"\x00") else "utf-16-le"

    return "utf-8"


def dumps(obj, app=None, **kwargs):
    """Serialize ``obj`` to a JSON-formatted string. If there is an
    app context pushed, use the current app's configured encoder
    (:attr:`~flask.Flask.json_encoder`), or fall back to the default
    :class:`JSONEncoder`.

    Takes the same arguments as the built-in :func:`json.dumps`, and
    does some extra configuration based on the application. If the
    simplejson package is installed, it is preferred.

    :param obj: Object to serialize to JSON.
    :param app: App instance to use to configure the JSON encoder.
        Uses ``current_app`` if not given, and falls back to the default
        encoder when not in an app context.
    :param kwargs: Extra arguments passed to :func:`json.dumps`.

    .. versionchanged:: 1.0.3

        ``app`` can be passed directly, rather than requiring an app
        context for configuration.
    """
    _dump_arg_defaults(kwargs, app=app)
    encoding = kwargs.pop("encoding", None)
    rv = _json.dumps(obj, **kwargs)
    if encoding is not None and isinstance(rv, text_type):
        rv = rv.encode(encoding)
    return rv


def dump(obj, fp, app=None, **kwargs):
    """Like :func:`dumps` but writes into a file object."""
    _dump_arg_defaults(kwargs, app=app)
    encoding = kwargs.pop("encoding", None)
    if encoding is not None:
        fp = _wrap_writer_for_text(fp, encoding)
    _json.dump(obj, fp, **kwargs)


def loads(s, app=None, **kwargs):
    """Deserialize an object from a JSON-formatted string ``s``. If
    there is an app context pushed, use the current app's configured
    decoder (:attr:`~flask.Flask.json_decoder`), or fall back to the
    default :class:`JSONDecoder`.

    Takes the same arguments as the built-in :func:`json.loads`, and
    does some extra configuration based on the application. If the
    simplejson package is installed, it is preferred.

    :param s: JSON string to deserialize.
    :param app: App instance to use to configure the JSON decoder.
        Uses ``current_app`` if not given, and falls back to the default
        encoder when not in an app context.
    :param kwargs: Extra arguments passed to :func:`json.dumps`.

    .. versionchanged:: 1.0.3

        ``app`` can be passed directly, rather than requiring an app
        context for configuration.
    """
    _load_arg_defaults(kwargs, app=app)
    if isinstance(s, bytes):
        encoding = kwargs.pop("encoding", None)
        if encoding is None:
            encoding = detect_encoding(s)
        s = s.decode(encoding)
    return _json.loads(s, **kwargs)


def load(fp, app=None, **kwargs):
    """Like :func:`loads` but reads from a file object."""
    _load_arg_defaults(kwargs, app=app)
    if not PY2:
        fp = _wrap_reader_for_text(fp, kwargs.pop("encoding", None) or "utf-8")
    return _json.load(fp, **kwargs)


def htmlsafe_dumps(obj, **kwargs):
    """Works exactly like :func:`dumps` but is safe for use in ``<script>``
    tags.  It accepts the same arguments and returns a JSON string.  Note that
    this is available in templates through the ``|tojson`` filter which will
    also mark the result as safe.  Due to how this function escapes certain
    characters this is safe even if used outside of ``<script>`` tags.

    The following characters are escaped in strings:

    -   ``<``
    -   ``>``
    -   ``&``
    -   ``'``

    This makes it safe to embed such strings in any place in HTML with the
    notable exception of double quoted attributes.  In that case single
    quote your attributes or HTML escape it in addition.

    .. versionchanged:: 0.10
       This function's return value is now always safe for HTML usage, even
       if outside of script tags or if used in XHTML.  This rule does not
       hold true when using this function in HTML attributes that are double
       quoted.  Always single quote attributes if you use the ``|tojson``
       filter.  Alternatively use ``|tojson|forceescape``.
    """
    rv = (
        dumps(obj, **kwargs)
        .replace(u"<", u"\\u003c")
        .replace(u">", u"\\u003e")
        .replace(u"&", u"\\u0026")
        .replace(u"'", u"\\u0027")
    )
    if not _slash_escape:
        rv = rv.replace("\\/", "/")
    return rv


def htmlsafe_dump(obj, fp, **kwargs):
    """Like :func:`htmlsafe_dumps` but writes into a file object."""
    fp.write(text_type(htmlsafe_dumps(obj, **kwargs)))


def jsonify(*args, **kwargs):
    """This function wraps :func:`dumps` to add a few enhancements that make
    life easier.  It turns the JSON output into a :class:`~flask.Response`
    object with the :mimetype:`application/json` mimetype.  For convenience, it
    also converts multiple arguments into an array or multiple keyword arguments
    into a dict.  This means that both ``jsonify(1,2,3)`` and
    ``jsonify([1,2,3])`` serialize to ``[1,2,3]``.

    For clarity, the JSON serialization behavior has the following differences
    from :func:`dumps`:

    1. Single argument: Passed straight through to :func:`dumps`.
    2. Multiple arguments: Converted to an array before being passed to
       :func:`dumps`.
    3. Multiple keyword arguments: Converted to a dict before being passed to
       :func:`dumps`.
    4. Both args and kwargs: Behavior undefined and will throw an exception.

    Example usage::

        from flask import jsonify

        @app.route('/_get_current_user')
        def get_current_user():
            return jsonify(username=g.user.username,
                           email=g.user.email,
                           id=g.user.id)

    This will send a JSON response like this to the browser::

        {
            "username": "admin",
            "email": "admin@localhost",
            "id": 42
        }


    .. versionchanged:: 0.11
       Added support for serializing top-level arrays. This introduces a
       security risk in ancient browsers. See :ref:`json-security` for details.

    This function's response will be pretty printed if the
    ``JSONIFY_PRETTYPRINT_REGULAR`` config parameter is set to True or the
    Flask app is running in debug mode. Compressed (not pretty) formatting
    currently means no indents and no spaces after separators.

    .. versionadded:: 0.2
    """

    indent = None
    separators = (",", ":")

    if current_app.config["JSONIFY_PRETTYPRINT_REGULAR"] or current_app.debug:
        indent = 2
        separators = (", ", ": ")

    if args and kwargs:
        raise TypeError("jsonify() behavior undefined when passed both args and kwargs")
    elif len(args) == 1:  # single args are passed directly to dumps()
        data = args[0]
    else:
        data = args or kwargs

    return current_app.response_class(
        dumps(data, indent=indent, separators=separators) + "\n",
        mimetype=current_app.config["JSONIFY_MIMETYPE"],
    )


def tojson_filter(obj, **kwargs):
    return Markup(htmlsafe_dumps(obj, **kwargs))



from commons.dto import DefaultHeaderDto  # noqa
from ._CashReportBodyDto import CashReportBodyDto  # noqa


class CashReportDto(object):
    def __init__(self, header, body):
        # type: (DefaultHeaderDto, CashReportBodyDto) -> None
        self.header = header
        self.body = body



from abc import ABCMeta, abstractmethod


class PaymentResponse(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def to_xml(self):
        # type: () -> str
        raise NotImplementedError()



class StoreGoals(object):
    def __init__(self, amount_goal, average_sale_goal, operator_sale_goal, item_goals):
        self.amount_goal = amount_goal
        self.average_sale_goal = average_sale_goal
        self.operator_sale_goal = operator_sale_goal
        self.sold_quantity = 0
        self.item_goals = item_goals



"""HTTP/1.1 client library

A backport of the Python 3.3 http/client.py module for python-future.

<intro stuff goes here>
<other stuff, too>

HTTPConnection goes through a number of "states", which define when a client
may legally make another request or fetch the response for a particular
request. This diagram details these state transitions:

    (null)
      |
      | HTTPConnection()
      v
    Idle
      |
      | putrequest()
      v
    Request-started
      |
      | ( putheader() )*  endheaders()
      v
    Request-sent
      |
      | response = getresponse()
      v
    Unread-response   [Response-headers-read]
      |\____________________
      |                     |
      | response.read()     | putrequest()
      v                     v
    Idle                  Req-started-unread-response
                     ______/|
                   /        |
   response.read() |        | ( putheader() )*  endheaders()
                   v        v
       Request-started    Req-sent-unread-response
                            |
                            | response.read()
                            v
                          Request-sent

This diagram presents the following rules:
  -- a second request may not be started until {response-headers-read}
  -- a response [object] cannot be retrieved until {request-sent}
  -- there is no differentiation between an unread response body and a
     partially read response body

Note: this enforcement is applied by the HTTPConnection class. The
      HTTPResponse class does not enforce this state machine, which
      implies sophisticated clients may accelerate the request/response
      pipeline. Caution should be taken, though: accelerating the states
      beyond the above pattern may imply knowledge of the server's
      connection-close behavior for certain requests. For example, it
      is impossible to tell whether the server will close the connection
      UNTIL the response headers have been read; this means that further
      requests cannot be placed into the pipeline until it is known that
      the server will NOT be closing the connection.

Logical State                  __state            __response
-------------                  -------            ----------
Idle                           _CS_IDLE           None
Request-started                _CS_REQ_STARTED    None
Request-sent                   _CS_REQ_SENT       None
Unread-response                _CS_IDLE           <response_class>
Req-started-unread-response    _CS_REQ_STARTED    <response_class>
Req-sent-unread-response       _CS_REQ_SENT       <response_class>
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from future.builtins import bytes, int, str, super
from future.utils import PY2

from future.backports.email import parser as email_parser
from future.backports.email import message as email_message
from future.backports.misc import create_connection as socket_create_connection
import io
import os
import socket
import collections
from future.backports.urllib.parse import urlsplit
import warnings
from array import array

__all__ = ["HTTPResponse", "HTTPConnection",
           "HTTPException", "NotConnected", "UnknownProtocol",
           "UnknownTransferEncoding", "UnimplementedFileMode",
           "IncompleteRead", "InvalidURL", "ImproperConnectionState",
           "CannotSendRequest", "CannotSendHeader", "ResponseNotReady",
           "BadStatusLine", "error", "responses"]

HTTP_PORT = 80
HTTPS_PORT = 443

_UNKNOWN = 'UNKNOWN'

# connection states
_CS_IDLE = 'Idle'
_CS_REQ_STARTED = 'Request-started'
_CS_REQ_SENT = 'Request-sent'

# status codes
# informational
CONTINUE = 100
SWITCHING_PROTOCOLS = 101
PROCESSING = 102

# successful
OK = 200
CREATED = 201
ACCEPTED = 202
NON_AUTHORITATIVE_INFORMATION = 203
NO_CONTENT = 204
RESET_CONTENT = 205
PARTIAL_CONTENT = 206
MULTI_STATUS = 207
IM_USED = 226

# redirection
MULTIPLE_CHOICES = 300
MOVED_PERMANENTLY = 301
FOUND = 302
SEE_OTHER = 303
NOT_MODIFIED = 304
USE_PROXY = 305
TEMPORARY_REDIRECT = 307

# client error
BAD_REQUEST = 400
UNAUTHORIZED = 401
PAYMENT_REQUIRED = 402
FORBIDDEN = 403
NOT_FOUND = 404
METHOD_NOT_ALLOWED = 405
NOT_ACCEPTABLE = 406
PROXY_AUTHENTICATION_REQUIRED = 407
REQUEST_TIMEOUT = 408
CONFLICT = 409
GONE = 410
LENGTH_REQUIRED = 411
PRECONDITION_FAILED = 412
REQUEST_ENTITY_TOO_LARGE = 413
REQUEST_URI_TOO_LONG = 414
UNSUPPORTED_MEDIA_TYPE = 415
REQUESTED_RANGE_NOT_SATISFIABLE = 416
EXPECTATION_FAILED = 417
UNPROCESSABLE_ENTITY = 422
LOCKED = 423
FAILED_DEPENDENCY = 424
UPGRADE_REQUIRED = 426
PRECONDITION_REQUIRED = 428
TOO_MANY_REQUESTS = 429
REQUEST_HEADER_FIELDS_TOO_LARGE = 431

# server error
INTERNAL_SERVER_ERROR = 500
NOT_IMPLEMENTED = 501
BAD_GATEWAY = 502
SERVICE_UNAVAILABLE = 503
GATEWAY_TIMEOUT = 504
HTTP_VERSION_NOT_SUPPORTED = 505
INSUFFICIENT_STORAGE = 507
NOT_EXTENDED = 510
NETWORK_AUTHENTICATION_REQUIRED = 511

# Mapping status codes to official W3C names
responses = {
    100: 'Continue',
    101: 'Switching Protocols',

    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',

    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    306: '(Unused)',
    307: 'Temporary Redirect',

    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    428: 'Precondition Required',
    429: 'Too Many Requests',
    431: 'Request Header Fields Too Large',

    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    511: 'Network Authentication Required',
}

# maximal amount of data to read at one time in _safe_read
MAXAMOUNT = 1048576

# maximal line length when calling readline().
_MAXLINE = 65536
_MAXHEADERS = 100


class HTTPMessage(email_message.Message):
    # XXX The only usage of this method is in
    # http.server.CGIHTTPRequestHandler.  Maybe move the code there so
    # that it doesn't need to be part of the public API.  The API has
    # never been defined so this could cause backwards compatibility
    # issues.

    def getallmatchingheaders(self, name):
        """Find all header lines matching a given header name.

        Look through the list of headers and find all lines matching a given
        header name (and their continuation lines).  A list of the lines is
        returned, without interpretation.  If the header does not occur, an
        empty list is returned.  If the header occurs multiple times, all
        occurrences are returned.  Case is not important in the header name.

        """
        name = name.lower() + ':'
        n = len(name)
        lst = []
        hit = 0
        for line in self.keys():
            if line[:n].lower() == name:
                hit = 1
            elif not line[:1].isspace():
                hit = 0
            if hit:
                lst.append(line)
        return lst

def parse_headers(fp, _class=HTTPMessage):
    """Parses only RFC2822 headers from a file pointer.

    email Parser wants to see strings rather than bytes.
    But a TextIOWrapper around self.rfile would buffer too many bytes
    from the stream, bytes which we later need to read as bytes.
    So we read the correct bytes here, as bytes, for email Parser
    to parse.

    """
    headers = []
    while True:
        line = fp.readline(_MAXLINE + 1)
        if len(line) > _MAXLINE:
            raise LineTooLong("header line")
        headers.append(line)
        if len(headers) > _MAXHEADERS:
            raise HTTPException("got more than %d headers" % _MAXHEADERS)
        if line in (b'\r\n', b'\n', b''):
            break
    hstring = bytes(b'').join(headers).decode('iso-8859-1')
    return email_parser.Parser(_class=_class).parsestr(hstring)


_strict_sentinel = object()

class HTTPResponse(io.RawIOBase):

    # See RFC 2616 sec 19.6 and RFC 1945 sec 6 for details.

    # The bytes from the socket object are iso-8859-1 strings.
    # See RFC 2616 sec 2.2 which notes an exception for MIME-encoded
    # text following RFC 2047.  The basic status line parsing only
    # accepts iso-8859-1.

    def __init__(self, sock, debuglevel=0, strict=_strict_sentinel, method=None, url=None):
        # If the response includes a content-length header, we need to
        # make sure that the client doesn't read more than the
        # specified number of bytes.  If it does, it will block until
        # the server times out and closes the connection.  This will
        # happen if a self.fp.read() is done (without a size) whether
        # self.fp is buffered or not.  So, no self.fp.read() by
        # clients unless they know what they are doing.
        self.fp = sock.makefile("rb")
        self.debuglevel = debuglevel
        if strict is not _strict_sentinel:
            warnings.warn("the 'strict' argument isn't supported anymore; "
                "http.client now always assumes HTTP/1.x compliant servers.",
                DeprecationWarning, 2)
        self._method = method

        # The HTTPResponse object is returned via urllib.  The clients
        # of http and urllib expect different attributes for the
        # headers.  headers is used here and supports urllib.  msg is
        # provided as a backwards compatibility layer for http
        # clients.

        self.headers = self.msg = None

        # from the Status-Line of the response
        self.version = _UNKNOWN # HTTP-Version
        self.status = _UNKNOWN  # Status-Code
        self.reason = _UNKNOWN  # Reason-Phrase

        self.chunked = _UNKNOWN         # is "chunked" being used?
        self.chunk_left = _UNKNOWN      # bytes left to read in current chunk
        self.length = _UNKNOWN          # number of bytes left in response
        self.will_close = _UNKNOWN      # conn will close at end of response

    def _read_status(self):
        line = str(self.fp.readline(_MAXLINE + 1), "iso-8859-1")
        if len(line) > _MAXLINE:
            raise LineTooLong("status line")
        if self.debuglevel > 0:
            print("reply:", repr(line))
        if not line:
            # Presumably, the server closed the connection before
            # sending a valid response.
            raise BadStatusLine(line)
        try:
            version, status, reason = line.split(None, 2)
        except ValueError:
            try:
                version, status = line.split(None, 1)
                reason = ""
            except ValueError:
                # empty version will cause next test to fail.
                version = ""
        if not version.startswith("HTTP/"):
            self._close_conn()
            raise BadStatusLine(line)

        # The status code is a three-digit number
        try:
            status = int(status)
            if status < 100 or status > 999:
                raise BadStatusLine(line)
        except ValueError:
            raise BadStatusLine(line)
        return version, status, reason

    def begin(self):
        if self.headers is not None:
            # we've already started reading the response
            return

        # read until we get a non-100 response
        while True:
            version, status, reason = self._read_status()
            if status != CONTINUE:
                break
            # skip the header from the 100 response
            while True:
                skip = self.fp.readline(_MAXLINE + 1)
                if len(skip) > _MAXLINE:
                    raise LineTooLong("header line")
                skip = skip.strip()
                if not skip:
                    break
                if self.debuglevel > 0:
                    print("header:", skip)

        self.code = self.status = status
        self.reason = reason.strip()
        if version in ("HTTP/1.0", "HTTP/0.9"):
            # Some servers might still return "0.9", treat it as 1.0 anyway
            self.version = 10
        elif version.startswith("HTTP/1."):
            self.version = 11   # use HTTP/1.1 code for HTTP/1.x where x>=1
        else:
            raise UnknownProtocol(version)

        self.headers = self.msg = parse_headers(self.fp)

        if self.debuglevel > 0:
            for hdr in self.headers:
                print("header:", hdr, end=" ")

        # are we using the chunked-style of transfer encoding?
        tr_enc = self.headers.get("transfer-encoding")
        if tr_enc and tr_enc.lower() == "chunked":
            self.chunked = True
            self.chunk_left = None
        else:
            self.chunked = False

        # will the connection close at the end of the response?
        self.will_close = self._check_close()

        # do we have a Content-Length?
        # NOTE: RFC 2616, S4.4, #3 says we ignore this if tr_enc is "chunked"
        self.length = None
        length = self.headers.get("content-length")

         # are we using the chunked-style of transfer encoding?
        tr_enc = self.headers.get("transfer-encoding")
        if length and not self.chunked:
            try:
                self.length = int(length)
            except ValueError:
                self.length = None
            else:
                if self.length < 0:  # ignore nonsensical negative lengths
                    self.length = None
        else:
            self.length = None

        # does the body have a fixed length? (of zero)
        if (status == NO_CONTENT or status == NOT_MODIFIED or
            100 <= status < 200 or      # 1xx codes
            self._method == "HEAD"):
            self.length = 0

        # if the connection remains open, and we aren't using chunked, and
        # a content-length was not provided, then assume that the connection
        # WILL close.
        if (not self.will_close and
            not self.chunked and
            self.length is None):
            self.will_close = True

    def _check_close(self):
        conn = self.headers.get("connection")
        if self.version == 11:
            # An HTTP/1.1 proxy is assumed to stay open unless
            # explicitly closed.
            conn = self.headers.get("connection")
            if conn and "close" in conn.lower():
                return True
            return False

        # Some HTTP/1.0 implementations have support for persistent
        # connections, using rules different than HTTP/1.1.

        # For older HTTP, Keep-Alive indicates persistent connection.
        if self.headers.get("keep-alive"):
            return False

        # At least Akamai returns a "Connection: Keep-Alive" header,
        # which was supposed to be sent by the client.
        if conn and "keep-alive" in conn.lower():
            return False

        # Proxy-Connection is a netscape hack.
        pconn = self.headers.get("proxy-connection")
        if pconn and "keep-alive" in pconn.lower():
            return False

        # otherwise, assume it will close
        return True

    def _close_conn(self):
        fp = self.fp
        self.fp = None
        fp.close()

    def close(self):
        super().close() # set "closed" flag
        if self.fp:
            self._close_conn()

    # These implementations are for the benefit of io.BufferedReader.

    # XXX This class should probably be revised to act more like
    # the "raw stream" that BufferedReader expects.

    def flush(self):
        super().flush()
        if self.fp:
            self.fp.flush()

    def readable(self):
        return True

    # End of "raw stream" methods

    def isclosed(self):
        """True if the connection is closed."""
        # NOTE: it is possible that we will not ever call self.close(). This
        #       case occurs when will_close is TRUE, length is None, and we
        #       read up to the last byte, but NOT past it.
        #
        # IMPLIES: if will_close is FALSE, then self.close() will ALWAYS be
        #          called, meaning self.isclosed() is meaningful.
        return self.fp is None

    def read(self, amt=None):
        if self.fp is None:
            return bytes(b"")

        if self._method == "HEAD":
            self._close_conn()
            return bytes(b"")

        if amt is not None:
            # Amount is given, so call base class version
            # (which is implemented in terms of self.readinto)
            return bytes(super(HTTPResponse, self).read(amt))
        else:
            # Amount is not given (unbounded read) so we must check self.length
            # and self.chunked

            if self.chunked:
                return self._readall_chunked()

            if self.length is None:
                s = self.fp.read()
            else:
                try:
                    s = self._safe_read(self.length)
                except IncompleteRead:
                    self._close_conn()
                    raise
                self.length = 0
            self._close_conn()        # we read everything
            return bytes(s)

    def readinto(self, b):
        if self.fp is None:
            return 0

        if self._method == "HEAD":
            self._close_conn()
            return 0

        if self.chunked:
            return self._readinto_chunked(b)

        if self.length is not None:
            if len(b) > self.length:
                # clip the read to the "end of response"
                b = memoryview(b)[0:self.length]

        # we do not use _safe_read() here because this may be a .will_close
        # connection, and the user is reading more bytes than will be provided
        # (for example, reading in 1k chunks)

        if PY2:
            data = self.fp.read(len(b))
            n = len(data)
            b[:n] = data
        else:
            n = self.fp.readinto(b)

        if not n and b:
            # Ideally, we would raise IncompleteRead if the content-length
            # wasn't satisfied, but it might break compatibility.
            self._close_conn()
        elif self.length is not None:
            self.length -= n
            if not self.length:
                self._close_conn()
        return n

    def _read_next_chunk_size(self):
        # Read the next chunk size from the file
        line = self.fp.readline(_MAXLINE + 1)
        if len(line) > _MAXLINE:
            raise LineTooLong("chunk size")
        i = line.find(b";")
        if i >= 0:
            line = line[:i] # strip chunk-extensions
        try:
            return int(line, 16)
        except ValueError:
            # close the connection as protocol synchronisation is
            # probably lost
            self._close_conn()
            raise

    def _read_and_discard_trailer(self):
        # read and discard trailer up to the CRLF terminator
        ### note: we shouldn't have any trailers!
        while True:
            line = self.fp.readline(_MAXLINE + 1)
            if len(line) > _MAXLINE:
                raise LineTooLong("trailer line")
            if not line:
                # a vanishingly small number of sites EOF without
                # sending the trailer
                break
            if line in (b'\r\n', b'\n', b''):
                break

    def _readall_chunked(self):
        assert self.chunked != _UNKNOWN
        chunk_left = self.chunk_left
        value = []
        while True:
            if chunk_left is None:
                try:
                    chunk_left = self._read_next_chunk_size()
                    if chunk_left == 0:
                        break
                except ValueError:
                    raise IncompleteRead(bytes(b'').join(value))
            value.append(self._safe_read(chunk_left))

            # we read the whole chunk, get another
            self._safe_read(2)      # toss the CRLF at the end of the chunk
            chunk_left = None

        self._read_and_discard_trailer()

        # we read everything; close the "file"
        self._close_conn()

        return bytes(b'').join(value)

    def _readinto_chunked(self, b):
        assert self.chunked != _UNKNOWN
        chunk_left = self.chunk_left

        total_bytes = 0
        mvb = memoryview(b)
        while True:
            if chunk_left is None:
                try:
                    chunk_left = self._read_next_chunk_size()
                    if chunk_left == 0:
                        break
                except ValueError:
                    raise IncompleteRead(bytes(b[0:total_bytes]))

            if len(mvb) < chunk_left:
                n = self._safe_readinto(mvb)
                self.chunk_left = chunk_left - n
                return total_bytes + n
            elif len(mvb) == chunk_left:
                n = self._safe_readinto(mvb)
                self._safe_read(2)  # toss the CRLF at the end of the chunk
                self.chunk_left = None
                return total_bytes + n
            else:
                temp_mvb = mvb[0:chunk_left]
                n = self._safe_readinto(temp_mvb)
                mvb = mvb[n:]
                total_bytes += n

            # we read the whole chunk, get another
            self._safe_read(2)      # toss the CRLF at the end of the chunk
            chunk_left = None

        self._read_and_discard_trailer()

        # we read everything; close the "file"
        self._close_conn()

        return total_bytes

    def _safe_read(self, amt):
        """Read the number of bytes requested, compensating for partial reads.

        Normally, we have a blocking socket, but a read() can be interrupted
        by a signal (resulting in a partial read).

        Note that we cannot distinguish between EOF and an interrupt when zero
        bytes have been read. IncompleteRead() will be raised in this
        situation.

        This function should be used when <amt> bytes "should" be present for
        reading. If the bytes are truly not available (due to EOF), then the
        IncompleteRead exception can be used to detect the problem.
        """
        s = []
        while amt > 0:
            chunk = self.fp.read(min(amt, MAXAMOUNT))
            if not chunk:
                raise IncompleteRead(bytes(b'').join(s), amt)
            s.append(chunk)
            amt -= len(chunk)
        return bytes(b"").join(s)

    def _safe_readinto(self, b):
        """Same as _safe_read, but for reading into a buffer."""
        total_bytes = 0
        mvb = memoryview(b)
        while total_bytes < len(b):
            if MAXAMOUNT < len(mvb):
                temp_mvb = mvb[0:MAXAMOUNT]
                n = self.fp.readinto(temp_mvb)
            else:
                n = self.fp.readinto(mvb)
            if not n:
                raise IncompleteRead(bytes(mvb[0:total_bytes]), len(b))
            mvb = mvb[n:]
            total_bytes += n
        return total_bytes

    def fileno(self):
        return self.fp.fileno()

    def getheader(self, name, default=None):
        if self.headers is None:
            raise ResponseNotReady()
        headers = self.headers.get_all(name) or default
        if isinstance(headers, str) or not hasattr(headers, '__iter__'):
            return headers
        else:
            return ', '.join(headers)

    def getheaders(self):
        """Return list of (header, value) tuples."""
        if self.headers is None:
            raise ResponseNotReady()
        return list(self.headers.items())

    # We override IOBase.__iter__ so that it doesn't check for closed-ness

    def __iter__(self):
        return self

    # For compatibility with old-style urllib responses.

    def info(self):
        return self.headers

    def geturl(self):
        return self.url

    def getcode(self):
        return self.status

class HTTPConnection(object):

    _http_vsn = 11
    _http_vsn_str = 'HTTP/1.1'

    response_class = HTTPResponse
    default_port = HTTP_PORT
    auto_open = 1
    debuglevel = 0

    def __init__(self, host, port=None, strict=_strict_sentinel,
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT, source_address=None):
        if strict is not _strict_sentinel:
            warnings.warn("the 'strict' argument isn't supported anymore; "
                "http.client now always assumes HTTP/1.x compliant servers.",
                DeprecationWarning, 2)
        self.timeout = timeout
        self.source_address = source_address
        self.sock = None
        self._buffer = []
        self.__response = None
        self.__state = _CS_IDLE
        self._method = None
        self._tunnel_host = None
        self._tunnel_port = None
        self._tunnel_headers = {}

        self._set_hostport(host, port)

    def set_tunnel(self, host, port=None, headers=None):
        """ Sets up the host and the port for the HTTP CONNECT Tunnelling.

        The headers argument should be a mapping of extra HTTP headers
        to send with the CONNECT request.
        """
        self._tunnel_host = host
        self._tunnel_port = port
        if headers:
            self._tunnel_headers = headers
        else:
            self._tunnel_headers.clear()

    def _set_hostport(self, host, port):
        if port is None:
            i = host.rfind(':')
            j = host.rfind(']')         # ipv6 addresses have [...]
            if i > j:
                try:
                    port = int(host[i+1:])
                except ValueError:
                    if host[i+1:] == "": # http://foo.com:/ == http://foo.com/
                        port = self.default_port
                    else:
                        raise InvalidURL("nonnumeric port: '%s'" % host[i+1:])
                host = host[:i]
            else:
                port = self.default_port
            if host and host[0] == '[' and host[-1] == ']':
                host = host[1:-1]
        self.host = host
        self.port = port

    def set_debuglevel(self, level):
        self.debuglevel = level

    def _tunnel(self):
        self._set_hostport(self._tunnel_host, self._tunnel_port)
        connect_str = "CONNECT %s:%d HTTP/1.0\r\n" % (self.host, self.port)
        connect_bytes = connect_str.encode("ascii")
        self.send(connect_bytes)
        for header, value in self._tunnel_headers.items():
            header_str = "%s: %s\r\n" % (header, value)
            header_bytes = header_str.encode("latin-1")
            self.send(header_bytes)
        self.send(bytes(b'\r\n'))

        response = self.response_class(self.sock, method=self._method)
        (version, code, message) = response._read_status()

        if code != 200:
            self.close()
            raise socket.error("Tunnel connection failed: %d %s" % (code,
                                                                    message.strip()))
        while True:
            line = response.fp.readline(_MAXLINE + 1)
            if len(line) > _MAXLINE:
                raise LineTooLong("header line")
            if not line:
                # for sites which EOF without sending a trailer
                break
            if line in (b'\r\n', b'\n', b''):
                break

    def connect(self):
        """Connect to the host and port specified in __init__."""
        self.sock = socket_create_connection((self.host,self.port),
                                             self.timeout, self.source_address)
        if self._tunnel_host:
            self._tunnel()

    def close(self):
        """Close the connection to the HTTP server."""
        if self.sock:
            self.sock.close()   # close it manually... there may be other refs
            self.sock = None
        if self.__response:
            self.__response.close()
            self.__response = None
        self.__state = _CS_IDLE

    def send(self, data):
        """Send `data' to the server.
        ``data`` can be a string object, a bytes object, an array object, a
        file-like object that supports a .read() method, or an iterable object.
        """

        if self.sock is None:
            if self.auto_open:
                self.connect()
            else:
                raise NotConnected()

        if self.debuglevel > 0:
            print("send:", repr(data))
        blocksize = 8192
        # Python 2.7 array objects have a read method which is incompatible
        # with the 2-arg calling syntax below.
        if hasattr(data, "read") and not isinstance(data, array):
            if self.debuglevel > 0:
                print("sendIng a read()able")
            encode = False
            try:
                mode = data.mode
            except AttributeError:
                # io.BytesIO and other file-like objects don't have a `mode`
                # attribute.
                pass
            else:
                if "b" not in mode:
                    encode = True
                    if self.debuglevel > 0:
                        print("encoding file using iso-8859-1")
            while 1:
                datablock = data.read(blocksize)
                if not datablock:
                    break
                if encode:
                    datablock = datablock.encode("iso-8859-1")
                self.sock.sendall(datablock)
            return
        try:
            self.sock.sendall(data)
        except TypeError:
            if isinstance(data, collections.Iterable):
                for d in data:
                    self.sock.sendall(d)
            else:
                raise TypeError("data should be a bytes-like object "
                                "or an iterable, got %r" % type(data))

    def _output(self, s):
        """Add a line of output to the current request buffer.

        Assumes that the line does *not* end with \\r\\n.
        """
        self._buffer.append(s)

    def _send_output(self, message_body=None):
        """Send the currently buffered request and clear the buffer.

        Appends an extra \\r\\n to the buffer.
        A message_body may be specified, to be appended to the request.
        """
        self._buffer.extend((bytes(b""), bytes(b"")))
        msg = bytes(b"\r\n").join(self._buffer)
        del self._buffer[:]
        # If msg and message_body are sent in a single send() call,
        # it will avoid performance problems caused by the interaction
        # between delayed ack and the Nagle algorithm.
        if isinstance(message_body, bytes):
            msg += message_body
            message_body = None
        self.send(msg)
        if message_body is not None:
            # message_body was not a string (i.e. it is a file), and
            # we must run the risk of Nagle.
            self.send(message_body)

    def putrequest(self, method, url, skip_host=0, skip_accept_encoding=0):
        """Send a request to the server.

        `method' specifies an HTTP request method, e.g. 'GET'.
        `url' specifies the object being requested, e.g. '/index.html'.
        `skip_host' if True does not add automatically a 'Host:' header
        `skip_accept_encoding' if True does not add automatically an
           'Accept-Encoding:' header
        """

        # if a prior response has been completed, then forget about it.
        if self.__response and self.__response.isclosed():
            self.__response = None


        # in certain cases, we cannot issue another request on this connection.
        # this occurs when:
        #   1) we are in the process of sending a request.   (_CS_REQ_STARTED)
        #   2) a response to a previous request has signalled that it is going
        #      to close the connection upon completion.
        #   3) the headers for the previous response have not been read, thus
        #      we cannot determine whether point (2) is true.   (_CS_REQ_SENT)
        #
        # if there is no prior response, then we can request at will.
        #
        # if point (2) is true, then we will have passed the socket to the
        # response (effectively meaning, "there is no prior response"), and
        # will open a new one when a new request is made.
        #
        # Note: if a prior response exists, then we *can* start a new request.
        #       We are not allowed to begin fetching the response to this new
        #       request, however, until that prior response is complete.
        #
        if self.__state == _CS_IDLE:
            self.__state = _CS_REQ_STARTED
        else:
            raise CannotSendRequest(self.__state)

        # Save the method we use, we need it later in the response phase
        self._method = method
        if not url:
            url = '/'
        request = '%s %s %s' % (method, url, self._http_vsn_str)

        # Non-ASCII characters should have been eliminated earlier
        self._output(request.encode('ascii'))

        if self._http_vsn == 11:
            # Issue some standard headers for better HTTP/1.1 compliance

            if not skip_host:
                # this header is issued *only* for HTTP/1.1
                # connections. more specifically, this means it is
                # only issued when the client uses the new
                # HTTPConnection() class. backwards-compat clients
                # will be using HTTP/1.0 and those clients may be
                # issuing this header themselves. we should NOT issue
                # it twice; some web servers (such as Apache) barf
                # when they see two Host: headers

                # If we need a non-standard port,include it in the
                # header.  If the request is going through a proxy,
                # but the host of the actual URL, not the host of the
                # proxy.

                netloc = ''
                if url.startswith('http'):
                    nil, netloc, nil, nil, nil = urlsplit(url)

                if netloc:
                    try:
                        netloc_enc = netloc.encode("ascii")
                    except UnicodeEncodeError:
                        netloc_enc = netloc.encode("idna")
                    self.putheader('Host', netloc_enc)
                else:
                    try:
                        host_enc = self.host.encode("ascii")
                    except UnicodeEncodeError:
                        host_enc = self.host.encode("idna")

                    # As per RFC 273, IPv6 address should be wrapped with []
                    # when used as Host header

                    if self.host.find(':') >= 0:
                        host_enc = bytes(b'[' + host_enc + b']')

                    if self.port == self.default_port:
                        self.putheader('Host', host_enc)
                    else:
                        host_enc = host_enc.decode("ascii")
                        self.putheader('Host', "%s:%s" % (host_enc, self.port))

            # note: we are assuming that clients will not attempt to set these
            #       headers since *this* library must deal with the
            #       consequences. this also means that when the supporting
            #       libraries are updated to recognize other forms, then this
            #       code should be changed (removed or updated).

            # we only want a Content-Encoding of "identity" since we don't
            # support encodings such as x-gzip or x-deflate.
            if not skip_accept_encoding:
                self.putheader('Accept-Encoding', 'identity')

            # we can accept "chunked" Transfer-Encodings, but no others
            # NOTE: no TE header implies *only* "chunked"
            #self.putheader('TE', 'chunked')

            # if TE is supplied in the header, then it must appear in a
            # Connection header.
            #self.putheader('Connection', 'TE')

        else:
            # For HTTP/1.0, the server will assume "not chunked"
            pass

    def putheader(self, header, *values):
        """Send a request header line to the server.

        For example: h.putheader('Accept', 'text/html')
        """
        if self.__state != _CS_REQ_STARTED:
            raise CannotSendHeader()

        if hasattr(header, 'encode'):
            header = header.encode('ascii')
        values = list(values)
        for i, one_value in enumerate(values):
            if hasattr(one_value, 'encode'):
                values[i] = one_value.encode('latin-1')
            elif isinstance(one_value, int):
                values[i] = str(one_value).encode('ascii')
        value = bytes(b'\r\n\t').join(values)
        header = header + bytes(b': ') + value
        self._output(header)

    def endheaders(self, message_body=None):
        """Indicate that the last header line has been sent to the server.

        This method sends the request to the server.  The optional message_body
        argument can be used to pass a message body associated with the
        request.  The message body will be sent in the same packet as the
        message headers if it is a string, otherwise it is sent as a separate
        packet.
        """
        if self.__state == _CS_REQ_STARTED:
            self.__state = _CS_REQ_SENT
        else:
            raise CannotSendHeader()
        self._send_output(message_body)

    def request(self, method, url, body=None, headers={}):
        """Send a complete request to the server."""
        self._send_request(method, url, body, headers)

    def _set_content_length(self, body):
        # Set the content-length based on the body.
        thelen = None
        try:
            thelen = str(len(body))
        except TypeError as te:
            # If this is a file-like object, try to
            # fstat its file descriptor
            try:
                thelen = str(os.fstat(body.fileno()).st_size)
            except (AttributeError, OSError):
                # Don't send a length if this failed
                if self.debuglevel > 0: print("Cannot stat!!")

        if thelen is not None:
            self.putheader('Content-Length', thelen)

    def _send_request(self, method, url, body, headers):
        # Honor explicitly requested Host: and Accept-Encoding: headers.
        header_names = dict.fromkeys([k.lower() for k in headers])
        skips = {}
        if 'host' in header_names:
            skips['skip_host'] = 1
        if 'accept-encoding' in header_names:
            skips['skip_accept_encoding'] = 1

        self.putrequest(method, url, **skips)

        if body is not None and ('content-length' not in header_names):
            self._set_content_length(body)
        for hdr, value in headers.items():
            self.putheader(hdr, value)
        if isinstance(body, str):
            # RFC 2616 Section 3.7.1 says that text default has a
            # default charset of iso-8859-1.
            body = body.encode('iso-8859-1')
        self.endheaders(body)

    def getresponse(self):
        """Get the response from the server.

        If the HTTPConnection is in the correct state, returns an
        instance of HTTPResponse or of whatever object is returned by
        class the response_class variable.

        If a request has not been sent or if a previous response has
        not be handled, ResponseNotReady is raised.  If the HTTP
        response indicates that the connection should be closed, then
        it will be closed before the response is returned.  When the
        connection is closed, the underlying socket is closed.
        """

        # if a prior response has been completed, then forget about it.
        if self.__response and self.__response.isclosed():
            self.__response = None

        # if a prior response exists, then it must be completed (otherwise, we
        # cannot read this response's header to determine the connection-close
        # behavior)
        #
        # note: if a prior response existed, but was connection-close, then the
        # socket and response were made independent of this HTTPConnection
        # object since a new request requires that we open a whole new
        # connection
        #
        # this means the prior response had one of two states:
        #   1) will_close: this connection was reset and the prior socket and
        #                  response operate independently
        #   2) persistent: the response was retained and we await its
        #                  isclosed() status to become true.
        #
        if self.__state != _CS_REQ_SENT or self.__response:
            raise ResponseNotReady(self.__state)

        if self.debuglevel > 0:
            response = self.response_class(self.sock, self.debuglevel,
                                           method=self._method)
        else:
            response = self.response_class(self.sock, method=self._method)

        response.begin()
        assert response.will_close != _UNKNOWN
        self.__state = _CS_IDLE

        if response.will_close:
            # this effectively passes the connection to the response
            self.close()
        else:
            # remember this, so we can tell when it is complete
            self.__response = response

        return response

try:
    import ssl
    from ssl import SSLContext
except ImportError:
    pass
else:
    class HTTPSConnection(HTTPConnection):
        "This class allows communication via SSL."

        default_port = HTTPS_PORT

        # XXX Should key_file and cert_file be deprecated in favour of context?

        def __init__(self, host, port=None, key_file=None, cert_file=None,
                     strict=_strict_sentinel, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                     source_address=None, **_3to2kwargs):
            if 'check_hostname' in _3to2kwargs: check_hostname = _3to2kwargs['check_hostname']; del _3to2kwargs['check_hostname']
            else: check_hostname = None
            if 'context' in _3to2kwargs: context = _3to2kwargs['context']; del _3to2kwargs['context']
            else: context = None
            super(HTTPSConnection, self).__init__(host, port, strict, timeout,
                                                  source_address)
            self.key_file = key_file
            self.cert_file = cert_file
            if context is None:
                # Some reasonable defaults
                context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                context.options |= ssl.OP_NO_SSLv2
            will_verify = context.verify_mode != ssl.CERT_NONE
            if check_hostname is None:
                check_hostname = will_verify
            elif check_hostname and not will_verify:
                raise ValueError("check_hostname needs a SSL context with "
                                 "either CERT_OPTIONAL or CERT_REQUIRED")
            if key_file or cert_file:
                context.load_cert_chain(cert_file, key_file)
            self._context = context
            self._check_hostname = check_hostname

        def connect(self):
            "Connect to a host on a given (SSL) port."

            sock = socket_create_connection((self.host, self.port),
                                            self.timeout, self.source_address)

            if self._tunnel_host:
                self.sock = sock
                self._tunnel()

            server_hostname = self.host if ssl.HAS_SNI else None
            self.sock = self._context.wrap_socket(sock,
                                                  server_hostname=server_hostname)
            try:
                if self._check_hostname:
                    ssl.match_hostname(self.sock.getpeercert(), self.host)
            except Exception:
                self.sock.shutdown(socket.SHUT_RDWR)
                self.sock.close()
                raise

    __all__.append("HTTPSConnection")


    # ######################################
    # # We use the old HTTPSConnection class from Py2.7, because ssl.SSLContext
    # # doesn't exist in the Py2.7 stdlib
    # class HTTPSConnection(HTTPConnection):
    #     "This class allows communication via SSL."

    #     default_port = HTTPS_PORT

    #     def __init__(self, host, port=None, key_file=None, cert_file=None,
    #                  strict=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
    #                  source_address=None):
    #         HTTPConnection.__init__(self, host, port, strict, timeout,
    #                                 source_address)
    #         self.key_file = key_file
    #         self.cert_file = cert_file

    #     def connect(self):
    #         "Connect to a host on a given (SSL) port."

    #         sock = socket_create_connection((self.host, self.port),
    #                                         self.timeout, self.source_address)
    #         if self._tunnel_host:
    #             self.sock = sock
    #             self._tunnel()
    #         self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file)

    # __all__.append("HTTPSConnection")
    # ######################################


class HTTPException(Exception):
    # Subclasses that define an __init__ must call Exception.__init__
    # or define self.args.  Otherwise, str() will fail.
    pass

class NotConnected(HTTPException):
    pass

class InvalidURL(HTTPException):
    pass

class UnknownProtocol(HTTPException):
    def __init__(self, version):
        self.args = version,
        self.version = version

class UnknownTransferEncoding(HTTPException):
    pass

class UnimplementedFileMode(HTTPException):
    pass

class IncompleteRead(HTTPException):
    def __init__(self, partial, expected=None):
        self.args = partial,
        self.partial = partial
        self.expected = expected
    def __repr__(self):
        if self.expected is not None:
            e = ', %i more expected' % self.expected
        else:
            e = ''
        return 'IncompleteRead(%i bytes read%s)' % (len(self.partial), e)
    def __str__(self):
        return repr(self)

class ImproperConnectionState(HTTPException):
    pass

class CannotSendRequest(ImproperConnectionState):
    pass

class CannotSendHeader(ImproperConnectionState):
    pass

class ResponseNotReady(ImproperConnectionState):
    pass

class BadStatusLine(HTTPException):
    def __init__(self, line):
        if not line:
            line = repr(line)
        self.args = line,
        self.line = line

class LineTooLong(HTTPException):
    def __init__(self, line_type):
        HTTPException.__init__(self, "got more than %d bytes when reading %s"
                                     % (_MAXLINE, line_type))

# for backwards compatibility
error = HTTPException



from datetime import datetime


class OrderSate(object):
    VOIDED = 4
    PAID = 5


class Order(object):
    def __init__(self, order_id, date, state):
        # type: (int, datetime, int) -> None
        self.order_id = order_id
        self.date = date
        self.state = state



from ._CancelPaymentOrderData import CancelPaymentOrderData
from ._ConfirmFinishPaymentData import ConfirmFinishPaymentData
from ._EftData import EftData
from ._FinishPaymentData import FinishPaymentData
from ._PaymentResponse import PaymentResponse
from ._PaymentResponseData import PaymentResponseData
from ._PaymentResponseMessage import PaymentResponseMessage
from ._StartPaymentData import StartPaymentData
from ._TestConnectionData import TestConnectionData
from ._CustomExceptions import (
    SitefUnavailableException,
    SitefClisitefDoesNotExistsException,
)



class InvalidTableEventException(RuntimeError):
    def __init__(self, data):
        message = 'Invalid Table Event. Event: "{}"'.format(data)
        super(InvalidTableEventException, self).__init__(message)



from httpmessageprocessorutil.automapped import auto_map
from typing import List


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "line_number": {"name": "lineNumber", "type": List[int]}})
class OrderDeleteLineRequest(object):
    def __init__(self, system_api_key, pos_id, line_numbers):
        # type: (str, str, List[int]) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.line_numbers = line_numbers



# -*- coding: utf-8 -*-


class InvalidJsonException(Exception):
    def __init__(self, message, invalid_json=None):
        # type: (unicode, unicode) -> None
        self.message = message
        self.invalid_json = invalid_json

    def __str__(self):
        # type: () -> unicode
        return u"InvalidJsonException: " + self.message + u" " + self.invalid_json

    def __repr__(self):
        # type: () -> unicode
        return self.__str__()



# encoding: utf-8
import time
from cStringIO import StringIO
from datetime import datetime
from xml.etree import cElementTree as eTree

from mw import MsgBusTableService as tableService
from old_helper import convert_from_utf_to_localtime
from utils import report_default_header, DATE_TIME_FMT, center, add_key_value_line, empty_line, add_table_line, \
    get_store_id, single_line_separator


class OpenedTablesReport(object):
    def __init__(self, table_repository, pos_id, operator_name_retriever):
        self.table_repository = table_repository
        self.pos_id = pos_id
        self.operator_name_retriever = operator_name_retriever
        self.store_id = get_store_id()

    def generate_opened_tables_report(self, pos_id):
        report = StringIO()
        title = "Relatorio Mesas Abertas"
        current_datetime = time.strftime(DATE_TIME_FMT)
        report.write(report_default_header(title, int(pos_id), None, None, current_datetime, self.store_id))
        self._opened_tables_report_body(report)

        return report.getvalue()

    def _opened_tables_report_body(self, report):
        tables = self.table_repository.get_opened_tables()

        columns_height = [18, 20]
        character_to_add = [".:", " ", " "]
        command = "Comanda"

        sectors_details = {}
        for table_id in tables:
            table_picture = tableService().get_table_picture(self.pos_id, str(table_id))
            table_picture_xml = eTree.XML(table_picture)
            table_orders = table_picture_xml.findall(".//Order")
            table_amount = sum([float(order.get("totalAmount")) for order in table_orders])
            date = datetime.strptime(table_picture_xml.get("startTS"), "%Y-%m-%d %H:%M:%S")
            date = convert_from_utf_to_localtime(date)
            operator_id = table_picture_xml.get("userId")
            operator_name = self.operator_name_retriever.get_operator_name(operator_id)
            seats = table_picture_xml.get("serviceSeats")

            columns = ["Total [Assentos]", 'R$ {:>8.2f}'.format(table_amount) + " [" + seats + "]"]
            is_tab = "TAB" in table_id
            table_context = "Comanda " if is_tab else "Mesa "

            table_name = table_picture_xml.get("name") if "name" in table_picture_xml.attrib else table_id
            table_properties = table_picture_xml.find("Properties")
            if table_properties is not None:
                table_name = table_properties.get("TAB_ID")[:7] if is_tab else table_name[:10]
            else:
                table_properties = table_picture_xml.find(".//OrderProperty[@key='TAB_ID']")
                table_name = table_properties.get("value")[:7] if is_tab else table_name[:10]

            table_detail_header = table_context + str(table_name) + " - " + datetime.strftime(date, "%d/%m/%Y %H:%M:%S")

            sector = command if is_tab else table_picture_xml.get("sector")
            table_info = dict(header=table_detail_header, operator=operator_name, columns=columns)
            if sector in sectors_details:
                sectors_details[sector].append(table_info)
            else:
                sectors_details[sector] = [table_info]

        sectors_keys = [s for s in sectors_details.keys() if s != command]
        sectors_keys.sort()
        if command in sectors_details.keys():
            sectors_keys.append(command)

        for sector in sectors_keys:
            report.write(center(sector.upper()))
            report.write(empty_line())
            for table_info in sectors_details[sector]:
                report.write(center(table_info["header"]))
                report.write(add_key_value_line("Operador........: ", table_info["operator"]))
                report.write(add_table_line(table_info["columns"], columns_height, character_to_add=character_to_add))
                report.write(empty_line())

            report.write(single_line_separator())



"""
A substitute for the Python 3 open() function.

Note that io.open() is more complete but maybe slower. Even so, the
completeness may be a better default. TODO: compare these
"""

_builtin_open = open

class newopen(object):
    """Wrapper providing key part of Python 3 open() interface.

    From IPython's py3compat.py module. License: BSD.
    """
    def __init__(self, fname, mode="r", encoding="utf-8"):
        self.f = _builtin_open(fname, mode)
        self.enc = encoding

    def write(self, s):
        return self.f.write(s.encode(self.enc))

    def read(self, size=-1):
        return self.f.read(size).decode(self.enc)

    def close(self):
        return self.f.close()

    def __enter__(self):
        return self

    def __exit__(self, etype, value, traceback):
        self.f.close()




from logging import Logger

from model.services import CouponServiceModel
from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import BaseRepository
from persistence import Connection
from sysactions import get_podtype, get_model
from typing import List, Any


class CouponService(CouponServiceModel, BaseRepository):

    def __init__(self, message_bus, logger):
        # type: (MbContextMessageBus, Logger) -> None

        super(CouponService, self).__init__(message_bus.mbcontext)
        self.message_bus = message_bus
        self.logger = logger
        self.electronic_payments_id = None

    def get_prices(self, posid):
        # type: (int) -> Any

        part_data = self._get_part_data()
        tag_data = self._get_tag_data()
        price_data = self._get_price_data(get_podtype(get_model(posid)) == "DL")

        prices = {}
        for product in price_data:
            if '.' not in product:
                prices[product] = price_data[product]

        def try_to_get_price(part_code, part_context):
            return price_data.get("{}.{}".format(part_context, part_code), price_data.get(part_code)) or 0

        def get_price_by_product_type_recursively(part_enters, produc):
            price = try_to_get_price(produc, '1')
            for part_params in part_enters:
                product_type = part_params["product_type"]
                part_code = part_params["part_code"]

                full_context = "{}.{}".format(produc, part_code)
                if product_type == 0:  # product
                    price += try_to_get_price(part_code, produc)
                elif product_type == 1:  # option
                    default_part = tag_data.get(produc.split('.')[-1], {}).get(part_code)
                    if default_part:
                        price += try_to_get_price(default_part, full_context)
                    else:
                        price += try_to_get_price(part_params["product_codes"][0], full_context)
                elif product_type == 2 and part_code in part_data:
                    price += get_price_by_product_type_recursively(part_data[part_code], full_context)
            return price

        for product, part_entries in part_data.iteritems():
            prices[product] = get_price_by_product_type_recursively(part_entries, product)

        return prices

    def _get_part_data(self):
        def inner_func(conn):
            # type: (Connection) -> List[int]

            part_data = {}
            cursor = conn.select("""
                                    SELECT
                                        part.productCode,
                                        part.partcode,
                                        part.defaultQty,
                                        params.ProductType,
                                        class.ProductCodes
                                    FROM ProductPart part
                                    JOIN ProductKernelParams params
                                        ON params.ProductCode = part.PartCode
                                    LEFT JOIN (
                                        SELECT
                                            classcode,
                                            group_concat(ProductCode) as ProductCodes
                                        FROM ProductClassification
                                        GROUP BY ClassCode
                                    ) class
                                         ON class.ClassCode = part.PartCode
                                    WHERE part.MinQty > 0
                                    ORDER BY part.ProductCode;
                                 """)
            for row in cursor:
                part_data.setdefault(row.get_entry(0), []).append({
                    "part_code": row.get_entry(1),
                    "default_qty": row.get_entry(2),
                    "product_type": int(row.get_entry(3)),
                    "product_codes": (row.get_entry(4) or "").split(',')
                })
            return part_data

        result_part_data = self.execute_with_connection(inner_func, service_name="Persistence")
        return result_part_data

    def _get_tag_data(self):
        def inner_func(conn):
            # type: (Connection) -> List[int]

            tag_data = {}
            cursor = conn.select("""
                        SELECT substr(Tag, 12), ProductCode
                        FROM ProductTags
                        WHERE Tag LIKE 'HasOptions=%'
                    """)
            for row in cursor:
                entry_data = {}
                for option_data in row.get_entry(0).split('|'):
                    option_code, default_option = option_data.split('>')
                    if not default_option:
                        continue
                    entry_data[option_code] = default_option
                tag_data[row.get_entry(1)] = entry_data
            return tag_data

        result_tag_data = self.execute_with_connection(inner_func, service_name="Persistence")
        return result_tag_data

    def _get_price_data(self, is_delivery):
        def inner_func(conn):
            # type: (Connection) -> List[int]

            price_data = {}
            cursor = conn.select("""
                    SELECT ProductCode, Context, COALESCE(CAST(DefaultUnitPrice AS FLOAT), 0)
                    FROM Price
                    WHERE PriceListId = '{}' AND DATETIME('now') BETWEEN ValidFrom AND ValidThru;
                    """.format("DL" if is_delivery else "EI"))
            for row in cursor:
                product_code, context, price = row
                if context and context != '1':
                    item_id = "{}.{}".format(context, product_code)
                else:
                    item_id = product_code
                price_data[item_id] = float(price)
            return price_data

        result_price_data = self.execute_with_connection(inner_func, service_name="Persistence")
        return result_price_data




class StaticConfig(object):
    def __init__(
        self,
        time_to_alert_table_opened,
        time_to_alert_table_is_idle,
        time_to_alert_table_is_idle_warning,
        time_to_alert_recall_delivery_is_idle,
        recall_button,
        operator_button,
        special_menus,
        enabled_tags,
        screen_timeout,
        cancel_timeout_window,
        show_in_dashboard,
        products_screen_dimensions,
        spinner_config,
        enable_tab_btns,
        enable_pre_start_sale,
        available_sale_types,
        fetch_stored_orders_timeout,
        tender_types,
        price_override_enabled,
        can_open_table_from_another_operator,
        can_edit_order,
        sat_info,
        navigation_config,
        show_cash_in_and_cash_out,
        show_void_closed_order,
        min_level_needed_to_see_all_tables,
        totem_config,
        cash_payment_enabled,
        coupons_enabled,
        discounts_enabled,
        bill_payment_enabled,
        special_modifiers,
        show_ruptured_products,
        pos_navigation,
        remote_order_status,
        delivery_sound,
        delivery_address,
        chat_enabled,
        services_period,
        menu_test,
        pre_account_with_discount,
        main_navigation,
        sort_table_order_desc,
        has_own_delivery_men,
        customer_info,
        table_service_mode,
        customer_info_tab,
        scramble_password_num_pad,
        store_id,
        min_pos_mobile_version
    ):
        self.time_to_alert_table_opened = time_to_alert_table_opened
        self.time_to_alert_table_is_idle = time_to_alert_table_is_idle
        self.time_to_alert_table_is_idle_warning = time_to_alert_table_is_idle_warning
        self.time_to_alert_recall_delivery_is_idle = time_to_alert_recall_delivery_is_idle
        self.recall_button = recall_button
        self.operator_button = operator_button
        self.special_menus = special_menus
        self.enabled_tags = enabled_tags
        self.screen_timeout = screen_timeout
        self.cancel_timeout_window = cancel_timeout_window
        self.show_in_dashboard = show_in_dashboard
        self.products_screen_dimensions = products_screen_dimensions
        self.spinner_config = spinner_config
        self.enable_tab_btns = enable_tab_btns
        self.enable_pre_start_sale = enable_pre_start_sale
        self.available_sale_types = available_sale_types
        self.fetch_stored_orders_timeout = fetch_stored_orders_timeout
        self.tender_types = tender_types
        self.price_override_enabled = price_override_enabled
        self.can_open_table_from_another_operator = can_open_table_from_another_operator
        self.can_edit_order = can_edit_order
        self.sat_info = sat_info
        self.navigation_config = navigation_config
        self.show_cash_in_and_cash_out = show_cash_in_and_cash_out
        self.show_void_closed_order = show_void_closed_order
        self.min_level_needed_to_see_all_tables = min_level_needed_to_see_all_tables
        self.totem_config = totem_config
        self.cash_payment_enabled = cash_payment_enabled
        self.coupons_enabled = coupons_enabled
        self.discounts_enabled = discounts_enabled
        self.bill_payment_enabled = bill_payment_enabled
        self.special_modifiers = special_modifiers
        self.show_ruptured_products = show_ruptured_products
        self.pos_navigation = pos_navigation
        self.remote_order_status = remote_order_status
        self.delivery_sound = delivery_sound
        self.delivery_address = delivery_address
        self.chat_enabled = chat_enabled
        self.services_period = services_period
        self.menu_test = menu_test
        self.pre_account_with_discount = pre_account_with_discount
        self.main_navigation = main_navigation
        self.sort_table_order_desc = sort_table_order_desc
        self.has_own_delivery_men = has_own_delivery_men
        self.customer_info = customer_info
        self.table_service_mode = table_service_mode
        self.customer_info_tab = customer_info_tab
        self.scramble_password_num_pad = scramble_password_num_pad
        self.store_id = store_id
        self.min_pos_mobile_version = min_pos_mobile_version



from model.electronic_message import ElectronicMessage
from typing import List


class ReadMessagesResponse(object):
    def __init__(self, messages):
        # type: (List[ElectronicMessage]) -> None

        self.messages = messages



# -*- coding: utf-8 -*-
"""
    flask.helpers
    ~~~~~~~~~~~~~

    Implements various helpers.

    :copyright: 2010 Pallets
    :license: BSD-3-Clause
"""
import io
import mimetypes
import os
import pkgutil
import posixpath
import socket
import sys
import unicodedata
from functools import update_wrapper
from threading import RLock
from time import time
from zlib import adler32

from jinja2 import FileSystemLoader
from werkzeug.datastructures import Headers
from werkzeug.exceptions import BadRequest
from werkzeug.exceptions import NotFound
from werkzeug.exceptions import RequestedRangeNotSatisfiable
from werkzeug.routing import BuildError
from werkzeug.urls import url_quote
from werkzeug.wsgi import wrap_file

from ._compat import fspath
from ._compat import PY2
from ._compat import string_types
from ._compat import text_type
from .globals import _app_ctx_stack
from .globals import _request_ctx_stack
from .globals import current_app
from .globals import request
from .globals import session
from .signals import message_flashed

# sentinel
_missing = object()


# what separators does this operating system provide that are not a slash?
# this is used by the send_from_directory function to ensure that nobody is
# able to access files from outside the filesystem.
_os_alt_seps = list(
    sep for sep in [os.path.sep, os.path.altsep] if sep not in (None, "/")
)


def get_env():
    """Get the environment the app is running in, indicated by the
    :envvar:`FLASK_ENV` environment variable. The default is
    ``'production'``.
    """
    return os.environ.get("FLASK_ENV") or "production"


def get_debug_flag():
    """Get whether debug mode should be enabled for the app, indicated
    by the :envvar:`FLASK_DEBUG` environment variable. The default is
    ``True`` if :func:`.get_env` returns ``'development'``, or ``False``
    otherwise.
    """
    val = os.environ.get("FLASK_DEBUG")

    if not val:
        return get_env() == "development"

    return val.lower() not in ("0", "false", "no")


def get_load_dotenv(default=True):
    """Get whether the user has disabled loading dotenv files by setting
    :envvar:`FLASK_SKIP_DOTENV`. The default is ``True``, load the
    files.

    :param default: What to return if the env var isn't set.
    """
    val = os.environ.get("FLASK_SKIP_DOTENV")

    if not val:
        return default

    return val.lower() in ("0", "false", "no")


def _endpoint_from_view_func(view_func):
    """Internal helper that returns the default endpoint for a given
    function.  This always is the function name.
    """
    assert view_func is not None, "expected view func if endpoint is not provided."
    return view_func.__name__


def stream_with_context(generator_or_function):
    """Request contexts disappear when the response is started on the server.
    This is done for efficiency reasons and to make it less likely to encounter
    memory leaks with badly written WSGI middlewares.  The downside is that if
    you are using streamed responses, the generator cannot access request bound
    information any more.

    This function however can help you keep the context around for longer::

        from flask import stream_with_context, request, Response

        @app.route('/stream')
        def streamed_response():
            @stream_with_context
            def generate():
                yield 'Hello '
                yield request.args['name']
                yield '!'
            return Response(generate())

    Alternatively it can also be used around a specific generator::

        from flask import stream_with_context, request, Response

        @app.route('/stream')
        def streamed_response():
            def generate():
                yield 'Hello '
                yield request.args['name']
                yield '!'
            return Response(stream_with_context(generate()))

    .. versionadded:: 0.9
    """
    try:
        gen = iter(generator_or_function)
    except TypeError:

        def decorator(*args, **kwargs):
            gen = generator_or_function(*args, **kwargs)
            return stream_with_context(gen)

        return update_wrapper(decorator, generator_or_function)

    def generator():
        ctx = _request_ctx_stack.top
        if ctx is None:
            raise RuntimeError(
                "Attempted to stream with context but "
                "there was no context in the first place to keep around."
            )
        with ctx:
            # Dummy sentinel.  Has to be inside the context block or we're
            # not actually keeping the context around.
            yield None

            # The try/finally is here so that if someone passes a WSGI level
            # iterator in we're still running the cleanup logic.  Generators
            # don't need that because they are closed on their destruction
            # automatically.
            try:
                for item in gen:
                    yield item
            finally:
                if hasattr(gen, "close"):
                    gen.close()

    # The trick is to start the generator.  Then the code execution runs until
    # the first dummy None is yielded at which point the context was already
    # pushed.  This item is discarded.  Then when the iteration continues the
    # real generator is executed.
    wrapped_g = generator()
    next(wrapped_g)
    return wrapped_g


def make_response(*args):
    """Sometimes it is necessary to set additional headers in a view.  Because
    views do not have to return response objects but can return a value that
    is converted into a response object by Flask itself, it becomes tricky to
    add headers to it.  This function can be called instead of using a return
    and you will get a response object which you can use to attach headers.

    If view looked like this and you want to add a new header::

        def index():
            return render_template('index.html', foo=42)

    You can now do something like this::

        def index():
            response = make_response(render_template('index.html', foo=42))
            response.headers['X-Parachutes'] = 'parachutes are cool'
            return response

    This function accepts the very same arguments you can return from a
    view function.  This for example creates a response with a 404 error
    code::

        response = make_response(render_template('not_found.html'), 404)

    The other use case of this function is to force the return value of a
    view function into a response which is helpful with view
    decorators::

        response = make_response(view_function())
        response.headers['X-Parachutes'] = 'parachutes are cool'

    Internally this function does the following things:

    -   if no arguments are passed, it creates a new response argument
    -   if one argument is passed, :meth:`flask.Flask.make_response`
        is invoked with it.
    -   if more than one argument is passed, the arguments are passed
        to the :meth:`flask.Flask.make_response` function as tuple.

    .. versionadded:: 0.6
    """
    if not args:
        return current_app.response_class()
    if len(args) == 1:
        args = args[0]
    return current_app.make_response(args)


def url_for(endpoint, **values):
    """Generates a URL to the given endpoint with the method provided.

    Variable arguments that are unknown to the target endpoint are appended
    to the generated URL as query arguments.  If the value of a query argument
    is ``None``, the whole pair is skipped.  In case blueprints are active
    you can shortcut references to the same blueprint by prefixing the
    local endpoint with a dot (``.``).

    This will reference the index function local to the current blueprint::

        url_for('.index')

    For more information, head over to the :ref:`Quickstart <url-building>`.

    Configuration values ``APPLICATION_ROOT`` and ``SERVER_NAME`` are only used when
    generating URLs outside of a request context.

    To integrate applications, :class:`Flask` has a hook to intercept URL build
    errors through :attr:`Flask.url_build_error_handlers`.  The `url_for`
    function results in a :exc:`~werkzeug.routing.BuildError` when the current
    app does not have a URL for the given endpoint and values.  When it does, the
    :data:`~flask.current_app` calls its :attr:`~Flask.url_build_error_handlers` if
    it is not ``None``, which can return a string to use as the result of
    `url_for` (instead of `url_for`'s default to raise the
    :exc:`~werkzeug.routing.BuildError` exception) or re-raise the exception.
    An example::

        def external_url_handler(error, endpoint, values):
            "Looks up an external URL when `url_for` cannot build a URL."
            # This is an example of hooking the build_error_handler.
            # Here, lookup_url is some utility function you've built
            # which looks up the endpoint in some external URL registry.
            url = lookup_url(endpoint, **values)
            if url is None:
                # External lookup did not have a URL.
                # Re-raise the BuildError, in context of original traceback.
                exc_type, exc_value, tb = sys.exc_info()
                if exc_value is error:
                    raise exc_type, exc_value, tb
                else:
                    raise error
            # url_for will use this result, instead of raising BuildError.
            return url

        app.url_build_error_handlers.append(external_url_handler)

    Here, `error` is the instance of :exc:`~werkzeug.routing.BuildError`, and
    `endpoint` and `values` are the arguments passed into `url_for`.  Note
    that this is for building URLs outside the current application, and not for
    handling 404 NotFound errors.

    .. versionadded:: 0.10
       The `_scheme` parameter was added.

    .. versionadded:: 0.9
       The `_anchor` and `_method` parameters were added.

    .. versionadded:: 0.9
       Calls :meth:`Flask.handle_build_error` on
       :exc:`~werkzeug.routing.BuildError`.

    :param endpoint: the endpoint of the URL (name of the function)
    :param values: the variable arguments of the URL rule
    :param _external: if set to ``True``, an absolute URL is generated. Server
      address can be changed via ``SERVER_NAME`` configuration variable which
      falls back to the `Host` header, then to the IP and port of the request.
    :param _scheme: a string specifying the desired URL scheme. The `_external`
      parameter must be set to ``True`` or a :exc:`ValueError` is raised. The default
      behavior uses the same scheme as the current request, or
      ``PREFERRED_URL_SCHEME`` from the :ref:`app configuration <config>` if no
      request context is available. As of Werkzeug 0.10, this also can be set
      to an empty string to build protocol-relative URLs.
    :param _anchor: if provided this is added as anchor to the URL.
    :param _method: if provided this explicitly specifies an HTTP method.
    """
    appctx = _app_ctx_stack.top
    reqctx = _request_ctx_stack.top

    if appctx is None:
        raise RuntimeError(
            "Attempted to generate a URL without the application context being"
            " pushed. This has to be executed when application context is"
            " available."
        )

    # If request specific information is available we have some extra
    # features that support "relative" URLs.
    if reqctx is not None:
        url_adapter = reqctx.url_adapter
        blueprint_name = request.blueprint

        if endpoint[:1] == ".":
            if blueprint_name is not None:
                endpoint = blueprint_name + endpoint
            else:
                endpoint = endpoint[1:]

        external = values.pop("_external", False)

    # Otherwise go with the url adapter from the appctx and make
    # the URLs external by default.
    else:
        url_adapter = appctx.url_adapter

        if url_adapter is None:
            raise RuntimeError(
                "Application was not able to create a URL adapter for request"
                " independent URL generation. You might be able to fix this by"
                " setting the SERVER_NAME config variable."
            )

        external = values.pop("_external", True)

    anchor = values.pop("_anchor", None)
    method = values.pop("_method", None)
    scheme = values.pop("_scheme", None)
    appctx.app.inject_url_defaults(endpoint, values)

    # This is not the best way to deal with this but currently the
    # underlying Werkzeug router does not support overriding the scheme on
    # a per build call basis.
    old_scheme = None
    if scheme is not None:
        if not external:
            raise ValueError("When specifying _scheme, _external must be True")
        old_scheme = url_adapter.url_scheme
        url_adapter.url_scheme = scheme

    try:
        try:
            rv = url_adapter.build(
                endpoint, values, method=method, force_external=external
            )
        finally:
            if old_scheme is not None:
                url_adapter.url_scheme = old_scheme
    except BuildError as error:
        # We need to inject the values again so that the app callback can
        # deal with that sort of stuff.
        values["_external"] = external
        values["_anchor"] = anchor
        values["_method"] = method
        values["_scheme"] = scheme
        return appctx.app.handle_url_build_error(error, endpoint, values)

    if anchor is not None:
        rv += "#" + url_quote(anchor)
    return rv


def get_template_attribute(template_name, attribute):
    """Loads a macro (or variable) a template exports.  This can be used to
    invoke a macro from within Python code.  If you for example have a
    template named :file:`_cider.html` with the following contents:

    .. sourcecode:: html+jinja

       {% macro hello(name) %}Hello {{ name }}!{% endmacro %}

    You can access this from Python code like this::

        hello = get_template_attribute('_cider.html', 'hello')
        return hello('World')

    .. versionadded:: 0.2

    :param template_name: the name of the template
    :param attribute: the name of the variable of macro to access
    """
    return getattr(current_app.jinja_env.get_template(template_name).module, attribute)


def flash(message, category="message"):
    """Flashes a message to the next request.  In order to remove the
    flashed message from the session and to display it to the user,
    the template has to call :func:`get_flashed_messages`.

    .. versionchanged:: 0.3
       `category` parameter added.

    :param message: the message to be flashed.
    :param category: the category for the message.  The following values
                     are recommended: ``'message'`` for any kind of message,
                     ``'error'`` for errors, ``'info'`` for information
                     messages and ``'warning'`` for warnings.  However any
                     kind of string can be used as category.
    """
    # Original implementation:
    #
    #     session.setdefault('_flashes', []).append((category, message))
    #
    # This assumed that changes made to mutable structures in the session are
    # always in sync with the session object, which is not true for session
    # implementations that use external storage for keeping their keys/values.
    flashes = session.get("_flashes", [])
    flashes.append((category, message))
    session["_flashes"] = flashes
    message_flashed.send(
        current_app._get_current_object(), message=message, category=category
    )


def get_flashed_messages(with_categories=False, category_filter=()):
    """Pulls all flashed messages from the session and returns them.
    Further calls in the same request to the function will return
    the same messages.  By default just the messages are returned,
    but when `with_categories` is set to ``True``, the return value will
    be a list of tuples in the form ``(category, message)`` instead.

    Filter the flashed messages to one or more categories by providing those
    categories in `category_filter`.  This allows rendering categories in
    separate html blocks.  The `with_categories` and `category_filter`
    arguments are distinct:

    * `with_categories` controls whether categories are returned with message
      text (``True`` gives a tuple, where ``False`` gives just the message text).
    * `category_filter` filters the messages down to only those matching the
      provided categories.

    See :ref:`message-flashing-pattern` for examples.

    .. versionchanged:: 0.3
       `with_categories` parameter added.

    .. versionchanged:: 0.9
        `category_filter` parameter added.

    :param with_categories: set to ``True`` to also receive categories.
    :param category_filter: whitelist of categories to limit return values
    """
    flashes = _request_ctx_stack.top.flashes
    if flashes is None:
        _request_ctx_stack.top.flashes = flashes = (
            session.pop("_flashes") if "_flashes" in session else []
        )
    if category_filter:
        flashes = list(filter(lambda f: f[0] in category_filter, flashes))
    if not with_categories:
        return [x[1] for x in flashes]
    return flashes


def send_file(
    filename_or_fp,
    mimetype=None,
    as_attachment=False,
    attachment_filename=None,
    add_etags=True,
    cache_timeout=None,
    conditional=False,
    last_modified=None,
):
    """Sends the contents of a file to the client.  This will use the
    most efficient method available and configured.  By default it will
    try to use the WSGI server's file_wrapper support.  Alternatively
    you can set the application's :attr:`~Flask.use_x_sendfile` attribute
    to ``True`` to directly emit an ``X-Sendfile`` header.  This however
    requires support of the underlying webserver for ``X-Sendfile``.

    By default it will try to guess the mimetype for you, but you can
    also explicitly provide one.  For extra security you probably want
    to send certain files as attachment (HTML for instance).  The mimetype
    guessing requires a `filename` or an `attachment_filename` to be
    provided.

    ETags will also be attached automatically if a `filename` is provided. You
    can turn this off by setting `add_etags=False`.

    If `conditional=True` and `filename` is provided, this method will try to
    upgrade the response stream to support range requests.  This will allow
    the request to be answered with partial content response.

    Please never pass filenames to this function from user sources;
    you should use :func:`send_from_directory` instead.

    .. versionadded:: 0.2

    .. versionadded:: 0.5
       The `add_etags`, `cache_timeout` and `conditional` parameters were
       added.  The default behavior is now to attach etags.

    .. versionchanged:: 0.7
       mimetype guessing and etag support for file objects was
       deprecated because it was unreliable.  Pass a filename if you are
       able to, otherwise attach an etag yourself.  This functionality
       will be removed in Flask 1.0

    .. versionchanged:: 0.9
       cache_timeout pulls its default from application config, when None.

    .. versionchanged:: 0.12
       The filename is no longer automatically inferred from file objects. If
       you want to use automatic mimetype and etag support, pass a filepath via
       `filename_or_fp` or `attachment_filename`.

    .. versionchanged:: 0.12
       The `attachment_filename` is preferred over `filename` for MIME-type
       detection.

    .. versionchanged:: 1.0
        UTF-8 filenames, as specified in `RFC 2231`_, are supported.

    .. _RFC 2231: https://tools.ietf.org/html/rfc2231#section-4

    .. versionchanged:: 1.0.3
        Filenames are encoded with ASCII instead of Latin-1 for broader
        compatibility with WSGI servers.

    .. versionchanged:: 1.1
        Filename may be a :class:`~os.PathLike` object.

    .. versionadded:: 1.1
        Partial content supports :class:`~io.BytesIO`.

    :param filename_or_fp: the filename of the file to send.
                           This is relative to the :attr:`~Flask.root_path`
                           if a relative path is specified.
                           Alternatively a file object might be provided in
                           which case ``X-Sendfile`` might not work and fall
                           back to the traditional method.  Make sure that the
                           file pointer is positioned at the start of data to
                           send before calling :func:`send_file`.
    :param mimetype: the mimetype of the file if provided. If a file path is
                     given, auto detection happens as fallback, otherwise an
                     error will be raised.
    :param as_attachment: set to ``True`` if you want to send this file with
                          a ``Content-Disposition: attachment`` header.
    :param attachment_filename: the filename for the attachment if it
                                differs from the file's filename.
    :param add_etags: set to ``False`` to disable attaching of etags.
    :param conditional: set to ``True`` to enable conditional responses.

    :param cache_timeout: the timeout in seconds for the headers. When ``None``
                          (default), this value is set by
                          :meth:`~Flask.get_send_file_max_age` of
                          :data:`~flask.current_app`.
    :param last_modified: set the ``Last-Modified`` header to this value,
        a :class:`~datetime.datetime` or timestamp.
        If a file was passed, this overrides its mtime.
    """
    mtime = None
    fsize = None

    if hasattr(filename_or_fp, "__fspath__"):
        filename_or_fp = fspath(filename_or_fp)

    if isinstance(filename_or_fp, string_types):
        filename = filename_or_fp
        if not os.path.isabs(filename):
            filename = os.path.join(current_app.root_path, filename)
        file = None
        if attachment_filename is None:
            attachment_filename = os.path.basename(filename)
    else:
        file = filename_or_fp
        filename = None

    if mimetype is None:
        if attachment_filename is not None:
            mimetype = (
                mimetypes.guess_type(attachment_filename)[0]
                or "application/octet-stream"
            )

        if mimetype is None:
            raise ValueError(
                "Unable to infer MIME-type because no filename is available. "
                "Please set either `attachment_filename`, pass a filepath to "
                "`filename_or_fp` or set your own MIME-type via `mimetype`."
            )

    headers = Headers()
    if as_attachment:
        if attachment_filename is None:
            raise TypeError("filename unavailable, required for sending as attachment")

        if not isinstance(attachment_filename, text_type):
            attachment_filename = attachment_filename.decode("utf-8")

        try:
            attachment_filename = attachment_filename.encode("ascii")
        except UnicodeEncodeError:
            filenames = {
                "filename": unicodedata.normalize("NFKD", attachment_filename).encode(
                    "ascii", "ignore"
                ),
                "filename*": "UTF-8''%s" % url_quote(attachment_filename, safe=b""),
            }
        else:
            filenames = {"filename": attachment_filename}

        headers.add("Content-Disposition", "attachment", **filenames)

    if current_app.use_x_sendfile and filename:
        if file is not None:
            file.close()
        headers["X-Sendfile"] = filename
        fsize = os.path.getsize(filename)
        headers["Content-Length"] = fsize
        data = None
    else:
        if file is None:
            file = open(filename, "rb")
            mtime = os.path.getmtime(filename)
            fsize = os.path.getsize(filename)
            headers["Content-Length"] = fsize
        elif isinstance(file, io.BytesIO):
            try:
                fsize = file.getbuffer().nbytes
            except AttributeError:
                # Python 2 doesn't have getbuffer
                fsize = len(file.getvalue())
            headers["Content-Length"] = fsize
        data = wrap_file(request.environ, file)

    rv = current_app.response_class(
        data, mimetype=mimetype, headers=headers, direct_passthrough=True
    )

    if last_modified is not None:
        rv.last_modified = last_modified
    elif mtime is not None:
        rv.last_modified = mtime

    rv.cache_control.public = True
    if cache_timeout is None:
        cache_timeout = current_app.get_send_file_max_age(filename)
    if cache_timeout is not None:
        rv.cache_control.max_age = cache_timeout
        rv.expires = int(time() + cache_timeout)

    if add_etags and filename is not None:
        from warnings import warn

        try:
            rv.set_etag(
                "%s-%s-%s"
                % (
                    os.path.getmtime(filename),
                    os.path.getsize(filename),
                    adler32(
                        filename.encode("utf-8")
                        if isinstance(filename, text_type)
                        else filename
                    )
                    & 0xFFFFFFFF,
                )
            )
        except OSError:
            warn(
                "Access %s failed, maybe it does not exist, so ignore etags in "
                "headers" % filename,
                stacklevel=2,
            )

    if conditional:
        try:
            rv = rv.make_conditional(request, accept_ranges=True, complete_length=fsize)
        except RequestedRangeNotSatisfiable:
            if file is not None:
                file.close()
            raise
        # make sure we don't send x-sendfile for servers that
        # ignore the 304 status code for x-sendfile.
        if rv.status_code == 304:
            rv.headers.pop("x-sendfile", None)
    return rv


def safe_join(directory, *pathnames):
    """Safely join `directory` and zero or more untrusted `pathnames`
    components.

    Example usage::

        @app.route('/wiki/<path:filename>')
        def wiki_page(filename):
            filename = safe_join(app.config['WIKI_FOLDER'], filename)
            with open(filename, 'rb') as fd:
                content = fd.read()  # Read and process the file content...

    :param directory: the trusted base directory.
    :param pathnames: the untrusted pathnames relative to that directory.
    :raises: :class:`~werkzeug.exceptions.NotFound` if one or more passed
            paths fall out of its boundaries.
    """

    parts = [directory]

    for filename in pathnames:
        if filename != "":
            filename = posixpath.normpath(filename)

        if (
            any(sep in filename for sep in _os_alt_seps)
            or os.path.isabs(filename)
            or filename == ".."
            or filename.startswith("../")
        ):
            raise NotFound()

        parts.append(filename)

    return posixpath.join(*parts)


def send_from_directory(directory, filename, **options):
    """Send a file from a given directory with :func:`send_file`.  This
    is a secure way to quickly expose static files from an upload folder
    or something similar.

    Example usage::

        @app.route('/uploads/<path:filename>')
        def download_file(filename):
            return send_from_directory(app.config['UPLOAD_FOLDER'],
                                       filename, as_attachment=True)

    .. admonition:: Sending files and Performance

       It is strongly recommended to activate either ``X-Sendfile`` support in
       your webserver or (if no authentication happens) to tell the webserver
       to serve files for the given path on its own without calling into the
       web application for improved performance.

    .. versionadded:: 0.5

    :param directory: the directory where all the files are stored.
    :param filename: the filename relative to that directory to
                     download.
    :param options: optional keyword arguments that are directly
                    forwarded to :func:`send_file`.
    """
    filename = fspath(filename)
    directory = fspath(directory)
    filename = safe_join(directory, filename)
    if not os.path.isabs(filename):
        filename = os.path.join(current_app.root_path, filename)
    try:
        if not os.path.isfile(filename):
            raise NotFound()
    except (TypeError, ValueError):
        raise BadRequest()
    options.setdefault("conditional", True)
    return send_file(filename, **options)


def get_root_path(import_name):
    """Returns the path to a package or cwd if that cannot be found.  This
    returns the path of a package or the folder that contains a module.

    Not to be confused with the package path returned by :func:`find_package`.
    """
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, "__file__"):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == "__main__":
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, "get_filename"):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, "__file__", None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError(
                "No root path can be found for the provided "
                'module "%s".  This can happen because the '
                "module came from an import hook that does "
                "not provide file name information or because "
                "it's a namespace package.  In this case "
                "the root path needs to be explicitly "
                "provided." % import_name
            )

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


def _matching_loader_thinks_module_is_package(loader, mod_name):
    """Given the loader that loaded a module and the module this function
    attempts to figure out if the given module is actually a package.
    """
    # If the loader can tell us if something is a package, we can
    # directly ask the loader.
    if hasattr(loader, "is_package"):
        return loader.is_package(mod_name)
    # importlib's namespace loaders do not have this functionality but
    # all the modules it loads are packages, so we can take advantage of
    # this information.
    elif (
        loader.__class__.__module__ == "_frozen_importlib"
        and loader.__class__.__name__ == "NamespaceLoader"
    ):
        return True
    # Otherwise we need to fail with an error that explains what went
    # wrong.
    raise AttributeError(
        (
            "%s.is_package() method is missing but is required by Flask of "
            "PEP 302 import hooks.  If you do not use import hooks and "
            "you encounter this error please file a bug against Flask."
        )
        % loader.__class__.__name__
    )


def _find_package_path(root_mod_name):
    """Find the path where the module's root exists in"""
    if sys.version_info >= (3, 4):
        import importlib.util

        try:
            spec = importlib.util.find_spec(root_mod_name)
            if spec is None:
                raise ValueError("not found")
        # ImportError: the machinery told us it does not exist
        # ValueError:
        #    - the module name was invalid
        #    - the module name is __main__
        #    - *we* raised `ValueError` due to `spec` being `None`
        except (ImportError, ValueError):
            pass  # handled below
        else:
            # namespace package
            if spec.origin in {"namespace", None}:
                return os.path.dirname(next(iter(spec.submodule_search_locations)))
            # a package (with __init__.py)
            elif spec.submodule_search_locations:
                return os.path.dirname(os.path.dirname(spec.origin))
            # just a normal module
            else:
                return os.path.dirname(spec.origin)

    # we were unable to find the `package_path` using PEP 451 loaders
    loader = pkgutil.get_loader(root_mod_name)
    if loader is None or root_mod_name == "__main__":
        # import name is not found, or interactive/main module
        return os.getcwd()
    else:
        # For .egg, zipimporter does not have get_filename until Python 2.7.
        if hasattr(loader, "get_filename"):
            filename = loader.get_filename(root_mod_name)
        elif hasattr(loader, "archive"):
            # zipimporter's loader.archive points to the .egg or .zip
            # archive filename is dropped in call to dirname below.
            filename = loader.archive
        else:
            # At least one loader is missing both get_filename and archive:
            # Google App Engine's HardenedModulesHook
            #
            # Fall back to imports.
            __import__(root_mod_name)
            filename = sys.modules[root_mod_name].__file__
        package_path = os.path.abspath(os.path.dirname(filename))

        # In case the root module is a package we need to chop of the
        # rightmost part.  This needs to go through a helper function
        # because of python 3.3 namespace packages.
        if _matching_loader_thinks_module_is_package(loader, root_mod_name):
            package_path = os.path.dirname(package_path)

    return package_path


def find_package(import_name):
    """Finds a package and returns the prefix (or None if the package is
    not installed) as well as the folder that contains the package or
    module as a tuple.  The package path returned is the module that would
    have to be added to the pythonpath in order to make it possible to
    import the module.  The prefix is the path below which a UNIX like
    folder structure exists (lib, share etc.).
    """
    root_mod_name, _, _ = import_name.partition(".")
    package_path = _find_package_path(root_mod_name)
    site_parent, site_folder = os.path.split(package_path)
    py_prefix = os.path.abspath(sys.prefix)
    if package_path.startswith(py_prefix):
        return py_prefix, package_path
    elif site_folder.lower() == "site-packages":
        parent, folder = os.path.split(site_parent)
        # Windows like installations
        if folder.lower() == "lib":
            base_dir = parent
        # UNIX like installations
        elif os.path.basename(parent).lower() == "lib":
            base_dir = os.path.dirname(parent)
        else:
            base_dir = site_parent
        return base_dir, package_path
    return None, package_path


class locked_cached_property(object):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value.  Works like the one in Werkzeug but has a lock for
    thread safety.
    """

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func
        self.lock = RLock()

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        with self.lock:
            value = obj.__dict__.get(self.__name__, _missing)
            if value is _missing:
                value = self.func(obj)
                obj.__dict__[self.__name__] = value
            return value


class _PackageBoundObject(object):
    #: The name of the package or module that this app belongs to. Do not
    #: change this once it is set by the constructor.
    import_name = None

    #: Location of the template files to be added to the template lookup.
    #: ``None`` if templates should not be added.
    template_folder = None

    #: Absolute path to the package on the filesystem. Used to look up
    #: resources contained in the package.
    root_path = None

    def __init__(self, import_name, template_folder=None, root_path=None):
        self.import_name = import_name
        self.template_folder = template_folder

        if root_path is None:
            root_path = get_root_path(self.import_name)

        self.root_path = root_path
        self._static_folder = None
        self._static_url_path = None

        # circular import
        from .cli import AppGroup

        #: The Click command group for registration of CLI commands
        #: on the application and associated blueprints. These commands
        #: are accessible via the :command:`flask` command once the
        #: application has been discovered and blueprints registered.
        self.cli = AppGroup()

    @property
    def static_folder(self):
        """The absolute path to the configured static folder."""
        if self._static_folder is not None:
            return os.path.join(self.root_path, self._static_folder)

    @static_folder.setter
    def static_folder(self, value):
        self._static_folder = value

    @property
    def static_url_path(self):
        """The URL prefix that the static route will be accessible from.

        If it was not configured during init, it is derived from
        :attr:`static_folder`.
        """
        if self._static_url_path is not None:
            return self._static_url_path

        if self.static_folder is not None:
            basename = os.path.basename(self.static_folder)
            return ("/" + basename).rstrip("/")

    @static_url_path.setter
    def static_url_path(self, value):
        if value is not None:
            value = value.rstrip("/")

        self._static_url_path = value

    @property
    def has_static_folder(self):
        """This is ``True`` if the package bound object's container has a
        folder for static files.

        .. versionadded:: 0.5
        """
        return self.static_folder is not None

    @locked_cached_property
    def jinja_loader(self):
        """The Jinja loader for this package bound object.

        .. versionadded:: 0.5
        """
        if self.template_folder is not None:
            return FileSystemLoader(os.path.join(self.root_path, self.template_folder))

    def get_send_file_max_age(self, filename):
        """Provides default cache_timeout for the :func:`send_file` functions.

        By default, this function returns ``SEND_FILE_MAX_AGE_DEFAULT`` from
        the configuration of :data:`~flask.current_app`.

        Static file functions such as :func:`send_from_directory` use this
        function, and :func:`send_file` calls this function on
        :data:`~flask.current_app` when the given cache_timeout is ``None``. If a
        cache_timeout is given in :func:`send_file`, that timeout is used;
        otherwise, this method is called.

        This allows subclasses to change the behavior when sending files based
        on the filename.  For example, to set the cache timeout for .js files
        to 60 seconds::

            class MyFlask(flask.Flask):
                def get_send_file_max_age(self, name):
                    if name.lower().endswith('.js'):
                        return 60
                    return flask.Flask.get_send_file_max_age(self, name)

        .. versionadded:: 0.9
        """
        return total_seconds(current_app.send_file_max_age_default)

    def send_static_file(self, filename):
        """Function used internally to send static files from the static
        folder to the browser.

        .. versionadded:: 0.5
        """
        if not self.has_static_folder:
            raise RuntimeError("No static folder for this object")
        # Ensure get_send_file_max_age is called in all cases.
        # Here, we ensure get_send_file_max_age is called for Blueprints.
        cache_timeout = self.get_send_file_max_age(filename)
        return send_from_directory(
            self.static_folder, filename, cache_timeout=cache_timeout
        )

    def open_resource(self, resource, mode="rb"):
        """Opens a resource from the application's resource folder.  To see
        how this works, consider the following folder structure::

            /myapplication.py
            /schema.sql
            /static
                /style.css
            /templates
                /layout.html
                /index.html

        If you want to open the :file:`schema.sql` file you would do the
        following::

            with app.open_resource('schema.sql') as f:
                contents = f.read()
                do_something_with(contents)

        :param resource: the name of the resource.  To access resources within
                         subfolders use forward slashes as separator.
        :param mode: Open file in this mode. Only reading is supported,
            valid values are "r" (or "rt") and "rb".
        """
        if mode not in {"r", "rt", "rb"}:
            raise ValueError("Resources can only be opened for reading")

        return open(os.path.join(self.root_path, resource), mode)


def total_seconds(td):
    """Returns the total seconds from a timedelta object.

    :param timedelta td: the timedelta to be converted in seconds

    :returns: number of seconds
    :rtype: int
    """
    return td.days * 60 * 60 * 24 + td.seconds


def is_ip(value):
    """Determine if the given string is an IP address.

    Python 2 on Windows doesn't provide ``inet_pton``, so this only
    checks IPv4 addresses in that environment.

    :param value: value to check
    :type value: str

    :return: True if string is an IP address
    :rtype: bool
    """
    if PY2 and os.name == "nt":
        try:
            socket.inet_aton(value)
            return True
        except socket.error:
            return False

    for family in (socket.AF_INET, socket.AF_INET6):
        try:
            socket.inet_pton(family, value)
        except socket.error:
            pass
        else:
            return True

    return False
